"""
Create Tasks from dictionaries

Note: All validation errors should raise DAGSpecInitializationError, this
allows the CLI to signal that this is a user's input error and hides the
traceback and only displays the error message
"""
from functools import partial
from copy import copy, deepcopy
from pathlib import Path
from collections.abc import MutableMapping, Mapping
import platform
from difflib import get_close_matches

from ploomber import tasks, products
from ploomber.util.util import _make_iterable
from ploomber.util import validate, dotted_path
from ploomber.tasks.taskgroup import TaskGroup
from ploomber import validators
from ploomber.exceptions import DAGSpecInitializationError
from ploomber.products._resources import resolve_resources
from ploomber.io import pretty_print

suffix2taskclass = {
    '.py': tasks.NotebookRunner,
    '.R': tasks.NotebookRunner,
    '.Rmd': tasks.NotebookRunner,
    '.r': tasks.NotebookRunner,
    '.ipynb': tasks.NotebookRunner,
    '.sql': tasks.SQLScript,
    '.sh': tasks.ShellScript
}


def _safe_suffix(product):
    try:
        return Path(product).suffix
    except Exception:
        return None


def _looks_like_path(s):
    system = platform.system()
    if system == 'Windows':
        return '\\' in s
    else:
        return '/' in s


def _extension_typo(extension, valid_extensions):
    if extension and valid_extensions:
        return get_close_matches(extension, valid_extensions)
    else:
        return None


def task_class_from_source_str(source_str, lazy_import, reload, product):
    """
    The source field in a DAG spec is a string. The actual value needed to
    instantiate the task depends on the task class, but to make task class
    optional, we try to guess the appropriate task here. If the source_str
    needs any pre-processing to pass it to the task constructor, it also
    happens here. If product is not None, it's also used to determine if
    a task is a SQLScript or SQLDump
    """
    try:
        extension = Path(source_str).suffix
    except Exception as e:
        raise DAGSpecInitializationError('Failed to initialize task '
                                         f'from source {source_str!r}') from e

    # we verify if this is a valid dotted path
    # if lazy load is set to true, just locate the module without importing it

    fn_checker = (
        dotted_path.locate_dotted_path_root if lazy_import is True else
        partial(dotted_path.load_dotted_path, raise_=True, reload=reload))

    if extension and extension in suffix2taskclass:
        if extension == '.sql':
            if _safe_suffix(product) in {'.csv', '.parquet'}:
                return tasks.SQLDump
            else:
                possibilities = _extension_typo(_safe_suffix(product),
                                                ['.csv', '.parquet'])
                if possibilities:
                    ext = possibilities[0]
                    raise DAGSpecInitializationError(
                        f'Error parsing task with source {source_str!r}: '
                        f'{_safe_suffix(product)!r} is not a valid product '
                        f'extension. Did you mean: {ext!r}?')

        return suffix2taskclass[extension]
    elif _looks_like_path(source_str):
        raise DAGSpecInitializationError(
            'Failed to determine task class for '
            f'source {source_str!r} (invalid '
            f'extension {extension!r}). Valid extensions '
            f'are: {pretty_print.iterable(suffix2taskclass)}')
    elif lazy_import == 'skip':
        # Anything that has not been caught before is treated as a
        # Python function, thus we return a PythonCallable
        return tasks.PythonCallable
    elif '.' in source_str:
        try:
            imported = fn_checker(source_str)
            error = None
        except Exception as e:
            imported = None
            error = e

        if imported is None:
            raise DAGSpecInitializationError(
                'Failed to determine task class for '
                f'source {source_str!r}: {error!s}. ')
        else:
            return tasks.PythonCallable
    else:
        raise DAGSpecInitializationError(
            f'Failed to determine task source {source_str!r}\nValid extensions'
            f' are: {pretty_print.iterable(suffix2taskclass)}\n'
            'You can also define functions as [module_name].[function_name]')


def task_class_from_spec(task_spec, lazy_import, reload):
    """
    Returns the class for the TaskSpec, if the spec already has the class
    name (str), it just returns the actual class object with such name,
    otherwise it tries to guess based on the source string
    """
    class_name = task_spec.get('class', None)

    if class_name:
        try:
            class_ = validators.string.validate_task_class_name(class_name)
        except Exception as e:
            msg = f'Error validating Task spec (class field): {e.args[0]}'
            e.args = (msg, )
            raise
    else:
        class_ = task_class_from_source_str(
            task_spec['source'],
            lazy_import,
            reload,
            task_spec.get('product'),
        )

    return class_


def _init_source_for_task_class(source_str, task_class, project_root,
                                lazy_import, make_absolute):
    """
    Initialize source. Loads dotted path to callable if a PythonCallable
    task, otherwise it returns a path
    """
    if task_class is tasks.PythonCallable:
        if lazy_import:
            return source_str
        else:
            return dotted_path.load_dotted_path(source_str)
    else:
        path = Path(source_str)

        # NOTE: there is some inconsistent behavior here. project_root
        # will be none if DAGSpec was initialized with a dictionary, hence
        # this won't resolve to absolute paths - this is a bit confusing.
        # maybe always convert to absolute?
        if project_root and not path.is_absolute() and make_absolute:
            return Path(project_root, source_str)
        else:
            return path


class TaskSpec(MutableMapping):
    """
    A TaskSpec converts dictionaries to Task objects. This class is not
    intended to be used directly, but through DAGSpec

    Parameters
    ----------
    data : dict
        The data that holds the spec information
    meta : dict
        The "meta" section information from the calling DAGSpec
    project_root : str or pathlib.Path
        The project root folder. Relative paths in "product" are so to this
        folder
    lazy_import : bool, default=False
        If False, sources are loaded when initializing the spec (e.g.
        a dotted path is imported, a source loaded using a SourceLoader
        is converted to a Placeholder object)
    reload : bool, default=False
        Reloads modules before importing dotted paths to detect code changes
        if the module has already been imported. Has no effect if
        lazy_import=True.
    """
    def __init__(self,
                 data,
                 meta,
                 project_root,
                 lazy_import=False,
                 reload=False):
        self.data = deepcopy(data)
        self.meta = deepcopy(meta)
        self.project_root = project_root
        self.lazy_import = lazy_import

        self.validate()

        source_loader = meta['source_loader']

        # initialize required elements
        self.data['class'] = task_class_from_spec(self.data, lazy_import,
                                                  reload)
        # preprocess source obj, at this point it will either be a Path if the
        # task requires a file or a callable if it's a PythonCallable task
        self.data['source'] = _init_source_for_task_class(
            self.data['source'],
            self.data['class'],
            self.project_root,
            lazy_import,
            # only make sources absolute paths when not using a source loader
            # otherwise keep them relative
            make_absolute=source_loader is None)

        is_path = isinstance(self.data['source'], Path)

        # check if we need to use the source_loader. we don't if the path is
        # relative because that doesn't make sense with a source_loader, and
        # this gives the user the ability to load some files that might
        # not be part of the source loader
        if source_loader and is_path and not self.data['source'].is_absolute():
            if lazy_import:
                self.data['source'] = source_loader.path_to(
                    self.data['source'])
            else:
                self.data['source'] = source_loader[self.data['source']]

    def validate(self):
        """
        Validates the data schema
        """
        if 'upstream' not in self.data:
            self.data['upstream'] = None

        if self.meta['extract_product']:
            required = {'source'}
        else:
            required = {'product', 'source'}

        validate.keys(valid=None,
                      passed=self.data,
                      required=required,
                      name=repr(self))

        if self.meta['extract_upstream'] and self.data.get('upstream'):
            raise DAGSpecInitializationError(
                'Error validating task "{}", if '
                'meta.extract_upstream is set to True, tasks '
                'should not have an "upstream" key'.format(self.data))

        if self.meta['extract_product'] and self.data.get('product'):
            raise DAGSpecInitializationError(
                'Error validating task "{}", if '
                'meta.extract_product is set to True, tasks '
                'should not have a "product" key'.format(self.data))

    def to_task(self, dag):
        """
        Convert the spec to a Task or TaskGroup and add it to the dag.
        Returns a (task, upstream) tuple with the Task instance and list of
        upstream dependencies (as described in the 'upstream' key, if any,
        empty if no 'upstream' key). If the spec has a 'grid' key, a TaskGroup
        instance instead

        Parameters
        ----------
        dag
            The DAG to add the task(s) to
        """
        data = copy(self.data)
        upstream = _make_iterable(data.pop('upstream'))

        if 'grid' in data:
            data_source_ = data["source"]
            data_source = str(data_source_ if not hasattr(
                data_source_, '__name__') else data_source_.__name__)

            if 'name' not in data:
                raise DAGSpecInitializationError(
                    f'Error initializing task with '
                    f'source {data_source!r}: '
                    'tasks with \'grid\' must have a \'name\'')

            task_class = data.pop('class')
            product_class = _find_product_class(task_class, data, self.meta)
            product = data.pop('product')
            name = data.pop('name')
            grid = data.pop('grid')

            # hooks
            on_render = data.pop('on_render', None)
            on_finish = data.pop('on_finish', None)
            on_failure = data.pop('on_failure', None)

            if on_render:
                on_render = dotted_path.DottedPath(on_render,
                                                   lazy_load=self.lazy_import)

            if on_finish:
                on_finish = dotted_path.DottedPath(on_finish,
                                                   lazy_load=self.lazy_import)

            if on_failure:
                on_failure = dotted_path.DottedPath(on_failure,
                                                    lazy_load=self.lazy_import)

            params = data.pop('params', None)

            return TaskGroup.from_grid(task_class=task_class,
                                       product_class=product_class,
                                       product_primitive=product,
                                       task_kwargs=data,
                                       dag=dag,
                                       name=name,
                                       grid=grid,
                                       resolve_relative_to=self.project_root,
                                       on_render=on_render,
                                       on_finish=on_finish,
                                       on_failure=on_failure,
                                       params=params), upstream
        else:
            return _init_task(data=data,
                              meta=self.meta,
                              project_root=self.project_root,
                              lazy_import=self.lazy_import,
                              dag=dag), upstream

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def __iter__(self):
        for e in self.data:
            yield e

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return '{}({!r})'.format(type(self).__name__, self.data)


def _init_task(data, meta, project_root, lazy_import, dag):
    """Initialize a single task from a dictionary spec
    """
    task_dict = copy(data)
    class_ = task_dict.pop('class')

    product = _init_product(task_dict,
                            meta,
                            class_,
                            project_root,
                            lazy_import=lazy_import)

    _init_client(task_dict, lazy_import=lazy_import)

    source = task_dict.pop('source')

    name = task_dict.pop('name', None)

    on_finish = task_dict.pop('on_finish', None)
    on_render = task_dict.pop('on_render', None)
    on_failure = task_dict.pop('on_failure', None)

    if 'serializer' in task_dict:
        task_dict['serializer'] = dotted_path.DottedPath(
            task_dict['serializer'], lazy_load=lazy_import)

    if 'unserializer' in task_dict:
        task_dict['unserializer'] = dotted_path.DottedPath(
            task_dict['unserializer'], lazy_load=lazy_import)

    # edge case: if using lazy_import, we should not check if the kernel
    # is installed. this is used when exporting to Argo/Airflow using
    # soopervisor, since the exporting process should not require to have
    # the ir kernel installed. The same applies when Airflow has to convert
    # the DAG, the Airflow environment shouldn't require the ir kernel
    if (class_ == tasks.NotebookRunner and lazy_import
            and 'check_if_kernel_installed' not in task_dict):
        task_dict['check_if_kernel_installed'] = False

    # make paths to resources absolute
    if 'params' in task_dict:
        task_dict['params'] = resolve_resources(task_dict['params'],
                                                relative_to=project_root)

    try:
        task = class_(source=source,
                      product=product,
                      name=name,
                      dag=dag,
                      **task_dict)
    except Exception as e:
        source_ = pretty_print.try_relative_path(source)
        msg = (f'Failed to initialize {class_.__name__} task with '
               f'source {source_!r}.')
        raise DAGSpecInitializationError(msg) from e

    if on_finish:
        task.on_finish = dotted_path.DottedPath(on_finish,
                                                lazy_load=lazy_import)

    if on_render:
        task.on_render = dotted_path.DottedPath(on_render,
                                                lazy_load=lazy_import)

    if on_failure:
        task.on_failure = dotted_path.DottedPath(on_failure,
                                                 lazy_load=lazy_import)

    return task


# FIXME: how do we make a default product client? use the task's client?
def _init_product(task_dict, meta, task_class, root_path, lazy_import):
    """
    Initialize product.

    Resolution logic order:
        task.product_class
        meta.{task_class}.product_default_class

    Current limitation: When there is more than one product, they all must
    be from the same class.
    """
    product_raw = task_dict.pop('product')

    # return if we already have a product
    if isinstance(product_raw, products.product.Product):
        return product_raw

    CLASS = _find_product_class(task_class, task_dict, meta)

    if 'product_client' in task_dict:
        dp = dotted_path.DottedPath(task_dict.pop('product_client'),
                                    lazy_load=lazy_import,
                                    allow_return_none=False)

        if lazy_import:
            client = dp
        else:
            client = dp()

        kwargs = {'client': client}
    else:
        kwargs = {}

    # determine the base path for the product (only relevant if product
    # is a File)
    relative_to = (Path(task_dict['source']).parent
                   if meta['product_relative_to_source'] else root_path)

    # initialize Product instance
    return try_product_init(CLASS, product_raw, relative_to, kwargs)


def _find_product_class(task_class, task_dict, meta):
    key = 'product_default_class.' + task_class.__name__
    meta_product_default_class = get_value_at(meta, key)

    if 'product_class' in task_dict:
        return validate_product_class_name(task_dict.pop('product_class'))
    elif meta_product_default_class:
        return validate_product_class_name(meta_product_default_class)
    else:
        raise DAGSpecInitializationError(
            f'Could not determine a product class for task: '
            f'{task_dict!r}. Add an explicit value in the '
            '"product_class"')


def try_product_init(class_, product_raw, relative_to, kwargs):
    """Initializes Product (or MetaProduct)

    Parameters
    ----------
    class_ : class
        Product class

    product_raw : str, list or dict
        The raw value as indicated by the user in the pipeline.yaml file. str
        if a single file, list if a SQL relation or dict if a MetaProduct

    relative_to : str
        Prefix for all relative paths (only applicable to File products)

    kwargs : dict
        Other kwargs to initialize product
    """
    if isinstance(product_raw, Mapping):
        return {
            key: _try_product_init(class_,
                                   resolve_if_file(value, relative_to, class_),
                                   kwargs)
            for key, value in product_raw.items()
        }
    else:
        path_to_source = resolve_if_file(product_raw, relative_to, class_)
        return _try_product_init(class_, path_to_source, kwargs)


def _try_product_init(class_, path_to_source, kwargs):
    """
    Try to initialize product, raises a chained exception if not possible.
    To provide more context.
    """
    try:
        return class_(path_to_source, **kwargs)
    except Exception as e:
        kwargs_msg = f' and keyword arguments: {kwargs!r}' if kwargs else ''
        raise DAGSpecInitializationError(
            f'Error initializing {class_.__name__} with source: '
            f'{path_to_source!r}' + kwargs_msg) from e


def validate_product_class_name(product_class_name):
    try:
        return validators.string.validate_product_class_name(
            product_class_name)
    except Exception as e:
        msg = ('Error validating Task spec (product_class field): '
               f'{e.args[0]}')
        e.args = (msg, )
        raise


def resolve_if_file(product_raw, relative_to, class_):
    """Resolve Product argument if it's a File to make it an absolute path
    """
    try:
        return _resolve_if_file(product_raw, relative_to, class_)
    except Exception as e:
        e.args = ('Error initializing File with argument '
                  f'{product_raw!r} ({e})', )
        raise


def _resolve_if_file(product_raw, relative_to, class_):
    """Resolve File argument to make it an absolute path
    """
    # not a file, nothing to do...
    if class_ != products.File:
        return product_raw
    # resolve...
    elif relative_to:
        # To keep things consistent, product relative paths are so to the
        # pipeline.yaml file (not to the current working directory). This is
        # important because there is no guarantee that the process calling
        # this will be at the pipeline.yaml location. One example is
        # when using the integration with Jupyter notebooks, each notebook
        # will set its working directory to the current parent.
        return str(Path(relative_to, product_raw).resolve())
    # no realtive_to argument, nothing to do...
    else:
        return Path(product_raw).resolve()


def _init_client(task_dict, lazy_import):
    if 'client' in task_dict:
        dp = dotted_path.DottedPath(task_dict.pop('client'),
                                    lazy_load=lazy_import,
                                    allow_return_none=False)

        if lazy_import:
            task_dict['client'] = dp
        else:
            task_dict['client'] = dp()


def get_value_at(d, dotted_path):
    current = d

    for key in dotted_path.split('.'):
        try:
            current = current[key]
        except KeyError:
            return None

    return current



from ploomber.cli import (
    cloud,
    build,
    plot,
    task,
    report,
    interact,
    status,
    examples,
    install,
    nb,
)

__all__ = [
    'cloud',
    'task',
    'plot',
    'build',
    'report',
    'interact',
    'status',
    'examples',
    'install',
    'nb',
]



import click

from ploomber.cli.parsers import CustomParser
from ploomber.cli.io import cli_endpoint
from ploomber.util.default import extract_name
from ploomber.dag.plot import choose_backend
from ploomber.telemetry import telemetry


@cli_endpoint
@telemetry.log_call('plot')
def main():
    parser = CustomParser(description='Plot a pipeline', prog='ploomber plot')
    with parser:
        parser.add_argument(
            '--output',
            '-o',
            help='Where to save the plot, defaults to pipeline.png',
            default=None)

        parser.add_argument('--backend',
                            '-b',
                            help=('How to generate the plot: d3 or '
                                  'pygraphviz. Using whatever is '
                                  'available by default'),
                            default=None)

        parser.add_argument('--include-products',
                            '-i',
                            help='Include task products',
                            action='store_true')

    dag, args = parser.load_from_entry_point_arg()

    if args.output is not None:
        output = args.output
    else:
        name = extract_name(args.entry_point)
        ext = 'png' if choose_backend(args.backend) == 'pygraphviz' else 'html'
        output = (f'pipeline.{ext}'
                  if name is None else f'pipeline.{name}.{ext}')

    dag.plot(output=output,
             backend=args.backend,
             include_products=args.include_products)

    click.echo(f'Plot saved at: {output}')



"""
Mixins to implement Task.load()
"""
from pathlib import Path


def _file_load(product, key=None, **kwargs):
    import pandas as pd

    if key:
        path = str(product[key])
    else:
        path = str(product)

    _extension2function = {
        '.csv': pd.read_csv,
        '.parquet': pd.read_parquet,
    }

    extension = Path(str(path)).suffix

    fn = _extension2function.get(extension)

    if fn is None:
        raise NotImplementedError(
            f'Cannot load file {path!r}. Files with extension '
            f'{extension!r} are currently unsupported, load manually by '
            'calling the appropriate unserializer function')

    return fn(path, **kwargs)


class FileLoaderMixin:
    """
    Load product as a pandas.DataFrame. Only .csv and .parquet files supported

    Parameters
    ----------
    key
        Key to use, if the task generates multiple products
    **kwargs
        Keyword arguments to pass to pandas.read_{csv, parquet} function
    """
    def load(self, key=None, **kwargs):
        return _file_load(self.product, key=key, **kwargs)



import warnings
from multiprocessing import Pool
import logging

from tqdm.auto import tqdm
from ploomber.executors.abc import Executor
from ploomber.executors import _format
from ploomber.exceptions import DAGBuildError, DAGBuildEarlyStop
from ploomber.messagecollector import (BuildExceptionsCollector,
                                       BuildWarningsCollector)
from ploomber.constants import TaskStatus


class Serial(Executor):
    """Executor than runs one task at a time

    Parameters
    ----------
    build_in_subprocess : bool, optional
        Determines whether tasks should be executed in a subprocess or in the
        current process. For pipelines with a lot of PythonCallables loading
        large objects such as pandas.DataFrame, this option is recommended as
        it guarantees that memory will be cleared up upon task execution.
        Defaults to True.

    catch_exceptions : bool, optional
        Whether to catch exceptions raised when building tasks and running
        hooks. If True, exceptions are collected and displayed at the end,
        downstream tasks of failed ones are aborted (not executed at all),
        If any task raises a DAGBuildEarlyStop exception, the final exception
        raised will be of such type. If False, no catching is done, on_failure
        won't be executed and task status will not be updated and tracebacks
        from build and hooks won't be logged. Setting of to False is only
        useful for debugging purposes.

    catch_warnings : bool, optional
        If True, the executor catches all warnings raised by tasks and
        displays them at the end of execution. If catch_exceptions is True
        and there is an error building the DAG, capture warnings are still
        shown before raising the collected exceptions.

    """
    def __init__(self,
                 build_in_subprocess=True,
                 catch_exceptions=True,
                 catch_warnings=True):
        self._logger = logging.getLogger(__name__)
        self._build_in_subprocess = build_in_subprocess
        self._catch_exceptions = catch_exceptions
        self._catch_warnings = catch_warnings

    def __repr__(self):
        return ('Serial(build_in_subprocess={}, '
                'catch_exceptions={}, catch_warnings={})'.format(
                    self._build_in_subprocess, self._catch_exceptions,
                    self._catch_warnings))

    def __call__(self, dag, show_progress):
        super().__call__(dag)

        exceptions_all = BuildExceptionsCollector()
        warnings_all = BuildWarningsCollector()
        task_reports = []

        task_kwargs = {'catch_exceptions': self._catch_exceptions}

        scheduled = [
            dag[t] for t in dag if dag[t].exec_status != TaskStatus.Skipped
        ]

        if show_progress:
            scheduled = tqdm(scheduled, total=len(scheduled))

        for t in scheduled:
            if t.exec_status == TaskStatus.Aborted:
                continue

            if show_progress:
                label = ('Building task'
                         if t.exec_status == TaskStatus.WaitingExecution else
                         'Downloading product')
                scheduled.set_description(f'{label} {t.name!r}')

            if self._build_in_subprocess:
                fn = LazyFunction(
                    build_in_subprocess, {
                        'task': t,
                        'build_kwargs': task_kwargs,
                        'reports_all': task_reports
                    }, t)
            else:
                fn = LazyFunction(
                    build_in_current_process, {
                        'task': t,
                        'build_kwargs': task_kwargs,
                        'reports_all': task_reports
                    }, t)

            if self._catch_warnings:
                fn = LazyFunction(fn=catch_warnings,
                                  kwargs={
                                      'fn': fn,
                                      'warnings_all': warnings_all
                                  },
                                  task=t)
            else:
                # NOTE: this isn't doing anything
                fn = LazyFunction(fn=pass_exceptions,
                                  kwargs={'fn': fn},
                                  task=t)

            if self._catch_exceptions:
                fn = LazyFunction(fn=catch_exceptions,
                                  kwargs={
                                      'fn': fn,
                                      'exceptions_all': exceptions_all
                                  },
                                  task=t)

            fn()

        # end of for loop

        if warnings_all and self._catch_warnings:
            # NOTE: maybe raise one by one to keep the warning type
            warnings.warn(str(warnings_all))

        if exceptions_all and self._catch_exceptions:
            early_stop = any(
                [isinstance(m.obj, DAGBuildEarlyStop) for m in exceptions_all])
            if early_stop:
                raise DAGBuildEarlyStop('Ealy stopping DAG execution, '
                                        'at least one of the tasks that '
                                        'failed raised a DAGBuildEarlyStop '
                                        'exception:\n{}'.format(
                                            str(exceptions_all)))
            else:
                raise DAGBuildError(str(exceptions_all))

        return task_reports

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._logger = logging.getLogger(__name__)


class LazyFunction:
    def __init__(self, fn, kwargs, task):
        self.fn = fn
        self.kwargs = kwargs
        self.task = task

    def __call__(self):
        return self.fn(**self.kwargs)


def catch_warnings(fn, warnings_all):
    # TODO: we need a try catch in case fn() raises an exception
    with warnings.catch_warnings(record=True) as warnings_current:
        warnings.simplefilter("ignore", DeprecationWarning)

        result = fn()

    if warnings_current:
        w = [str(a_warning.message) for a_warning in warnings_current]
        warnings_all.append(task=fn.task, message='\n'.join(w))

    return result


def catch_exceptions(fn, exceptions_all):
    logger = logging.getLogger(__name__)
    # NOTE: we are individually catching exceptions
    # (see build_in_current_process and build_in_subprocess), would it be
    # better to catch everything here and set
    # task.exec_status = TaskStatus.Errored accordingly?

    # TODO: setting exec_status can also raise exceptions if the hook fails
    # add tests for that, and check the final task status,
    try:
        # try to run task build
        fn()
    except Exception as e:
        # if running in a different process, logger.exception inside Task.build
        # won't show up. So we do it here.
        # FIXME: this is going to cause duplicates if not running in a
        # subprocess
        logger.exception(str(e))
        tr = _format.exception(e)
        exceptions_all.append(task=fn.task, message=tr, obj=e)


def pass_exceptions(fn):
    # should i still check here for DAGBuildEarlyStop? is it worth
    # for returning accurate task status?
    fn()


def build_in_current_process(task, build_kwargs, reports_all):
    report, meta = task._build(**build_kwargs)
    reports_all.append(report)


def build_in_subprocess(task, build_kwargs, reports_all):
    if callable(task.source.primitive):

        try:
            p = Pool(processes=1)
        except RuntimeError as e:
            if 'An attempt has been made to start a new process' in str(e):
                # this is most likely due to child processes created with
                # spawn (mac/windows) outside if __name__ == '__main__'
                raise RuntimeError('Press ctrl + c to exit. '
                                   'For help solving this, go to: '
                                   'https://ploomber.io/s/mp') from e
            else:
                raise

        res = p.apply_async(func=task._build, kwds=build_kwargs)
        # calling this make sure we catch the exception, from the docs:
        # Return the result when it arrives. If timeout is not None and
        # the result does not arrive within timeout seconds then
        # multiprocessing.TimeoutError is raised. If the remote call
        # raised an exception then that exception will be reraised by
        # get().
        # https://docs.python.org/3/library/multiprocessing.html#multiprocessing.pool.AsyncResult.get
        try:
            report, meta = res.get()
        except Exception:
            # we have to update status since this is ran in a subprocess
            task.exec_status = TaskStatus.Errored
            raise
        else:
            task.product.metadata.update_locally(meta)
            task.exec_status = TaskStatus.Executed
            reports_all.append(report)
        finally:
            p.close()
            p.join()

    else:
        # we run other tasks in the same process
        report, meta = task._build(**build_kwargs)
        task.product.metadata.update_locally(meta)
        reports_all.append(report)



from jinja2 import Environment, meta
from jinja2.nodes import Assign

from ploomber.placeholders import extensions

extensions = (extensions.RaiseExtension, )
env_render = Environment(extensions=extensions)
env_runtime = Environment(extensions=extensions,
                          variable_start_string='[[',
                          variable_end_string=']]')


def get_tags_in_str(s, require_runtime_placeholders=True):
    """
    Returns tags (e.g. {{variable}}) in a given string as a set, returns an
    empty set for None

    Parameters
    ----------
    require_runtime_placeholders : bool, default=True
        Also check runtime tags - the ones in square brackets
        (e.g. [[placeholder]])
    """
    # render placeholders
    vars_render = meta.find_undeclared_variables(env_render.parse(s))

    # runtime placeholders
    if require_runtime_placeholders:
        vars_runtime = meta.find_undeclared_variables(env_runtime.parse(s))
    else:
        vars_runtime = set()

    return vars_render | vars_runtime


def get_defined_variables(s):
    env = Environment()
    ast = env.parse(s)
    return {n.target.name: n.node.as_const() for n in ast.find_all(Assign)}



"""
Extensions for the inspect module
"""
import inspect


def getfile(fn):
    """
    Returns the file where the function is defined. Works even in wrapped
    functions
    """
    if hasattr(fn, '__wrapped__'):
        return getfile(fn.__wrapped__)
    else:
        return inspect.getfile(fn)



from ploomber.io import pretty_print

from collections.abc import Mapping, Iterable
from itertools import product, chain


class Interval:
    """Generate intervals from lower to upper of size delta, last interval
    will be smaller if needed

    >>> from ploomber.util import Interval
    >>> from datetime import date
    >>> from dateutil.relativedelta import relativedelta
    >>> interval = Interval(date(year=2020, month=1, day=1),
    ...                     date(year=2022, month=6, day=1),
    ...                     relativedelta(years=1)).expand()
    >>> interval[0]
    (datetime.date(2020, 1, 1), datetime.date(2021, 1, 1))
    >>> interval[1]
    (datetime.date(2021, 1, 1), datetime.date(2022, 1, 1))
    >>> interval[2]
    (datetime.date(2022, 1, 1), datetime.date(2022, 6, 1))
    """
    def __init__(self, lower, upper, delta):
        if lower >= upper:
            raise ValueError('lower must be strictly lower than upper')

        self.lower = lower
        self.upper = upper
        self.delta = delta

    def expand(self):
        tuples = []
        cursor = self.lower
        i = 0

        while True:
            cursor = self.lower + (i + 1) * self.delta

            if cursor < self.upper:
                tuples.append((self.lower + i * self.delta, cursor))
                i += 1
            else:
                tuples.append((self.lower + i * self.delta, self.upper))
                break

        return tuples

    def __repr__(self):
        return ('Interval from {} to {} with delta {}'.format(
            self.lower, self.upper, self.delta))


class ParamGrid:
    """Generate parameter grids

    Parameters
    ----------
    grid : list or dict
        Grid to generate. Can generate multiple grids at once by passing
        a list of grids

    params : dict
        Parameters that should remain fixed

    Examples
    --------
    >>> pg = ParamGrid({'a': [1, 2, 3], 'b': [2, 4, 6]})
    >>> list(pg.zip())
    [{'a': 1, 'b': 2}, {'a': 2, 'b': 4}, {'a': 3, 'b': 6}]

    >>> list(pg.product())  # doctest: +NORMALIZE_WHITESPACE
    [{'a': 1, 'b': 2}, {'a': 1, 'b': 4}, {'a': 1, 'b': 6}, {'a': 2, 'b': 2},
    {'a': 2, 'b': 4}, {'a': 2, 'b': 6}, {'a': 3, 'b': 2}, {'a': 3, 'b': 4},
    {'a': 3, 'b': 6}]

    >>> pg = ParamGrid({'a': Interval(0, 10, 2), 'b': [2, 4, 6, 8, 10]})
    >>> list(pg.zip())  # doctest: +NORMALIZE_WHITESPACE
    [{'a': (0, 2), 'b': 2}, {'a': (2, 4), 'b': 4}, {'a': (4, 6), 'b': 6},
    {'a': (6, 8), 'b': 8}, {'a': (8, 10), 'b': 10}]

    Notes
    -----
    Parameters with a single element are converted to lists of length 1
    """
    def __init__(self, grid, params=None):
        if isinstance(grid, Mapping):
            grid = [grid]

        self._expanded = [_expand(d) for d in grid]
        self._params = params or dict()

    def zip(self):
        for d in chain(self._expanded):
            lengths = set(len(v) for v in d.values())

            if len(lengths) != 1:
                raise ValueError('All parameters should have the same length')

            length = list(lengths)[0]

            for i in range(length):
                out = {k: v[i] for k, v in d.items()}
                _check_keys_overlap(out, self._params)
                yield {**out, **self._params}

    def product(self):
        for d in chain(self._expanded):
            keys = d.keys()
            values = d.values()

            for elements in product(*values):
                d = {}

                for k, v in zip(keys, elements):
                    d[k] = v

                _check_keys_overlap(d, self._params)

                yield {**d, **self._params}


def _expand(d):
    expanded = {}

    for k, v in d.items():
        if isinstance(v, Interval):
            expanded[k] = v.expand()
        elif not isinstance(v, Iterable) or isinstance(v, str):
            expanded[k] = [v]
        else:
            expanded[k] = v

    return expanded


def _check_keys_overlap(a, b):

    overlap = set(a) & set(b)

    if overlap:
        overlap_ = pretty_print.iterable(overlap)
        raise ValueError("Error generating grid: 'grid' and 'params' "
                         f"have overlapping keys: {overlap_}")



import os
import sys
import subprocess
import shutil
from pathlib import Path, PurePosixPath

from click import ClickException
from jinja2 import Environment, PackageLoader, StrictUndefined

from ploomber.io import TerminalWriter


def to_pascal_case(name):
    return ''.join([w.capitalize() for w in name.split('_')])


def _delete(dst):
    dst = Path(dst)

    if dst.is_file():
        dst.unlink()

    if dst.is_dir():
        shutil.rmtree(dst)


class CommanderException(ClickException):
    """
    Exception raised when the workflow cannot proceed and require a fix
    from the user. It is a subclass of ClickException, which signals the CLI
    to hide the traceback
    """
    pass


class CommanderStop(Exception):
    """
    An exception that stops the execution of a commander without raising
    an exception
    """
    pass


class Commander:
    """Manage script workflows
    """
    def __init__(self,
                 workspace=None,
                 templates_path=None,
                 environment_kwargs=None):
        self.tw = TerminalWriter()
        self.workspace = None if not workspace else Path(workspace).resolve()
        self._to_delete = []
        self._warnings = []

        self._wd = Path('.').resolve()

        if templates_path:
            self._env = Environment(loader=PackageLoader(*templates_path),
                                    undefined=StrictUndefined,
                                    **(environment_kwargs or {}))
            self._env.filters['to_pascal_case'] = to_pascal_case
        else:
            self._env = None

    def run(self,
            *cmd,
            description=None,
            capture_output=False,
            expected_output=None,
            error_message=None,
            hint=None,
            show_cmd=True):
        """Execute a command in a subprocess

        Parameters
        ----------
        *cmd
            Command to execute

        description: str, default=None
            Label to display before executing the command

        capture_output: bool, default=False
            Captures output, otherwise prints to standard output and standard
            error

        expected_output: str, default=None
            Raises a RuntimeError if the output is different than this value.
            Only valid when capture_output=True

        error_message: str, default=None
            Error to display when expected_output does not match. If None,
            a generic message is shown

        hint: str, default=None
            An optional string to show when at the end of the error when
            the expected_output does not match. Used to hint the user how
            to fix the problem

        show_cmd : bool, default=True
            Whether to display the command next to the description
            (and error message if it fails) or not. Only valid when
            description is not None
        """
        cmd_str = ' '.join(cmd)

        if expected_output is not None and not capture_output:
            raise RuntimeError('capture_output must be True when '
                               'expected_output is not None')

        if description:
            header = f'{description}: {cmd_str}' if show_cmd else description
            self.tw.sep('=', header, blue=True)

        error = None

        # py 3.6 compatibility: cannot use subprocess.run directly
        # because the check_output arg was included until version 3.7
        if not capture_output:
            try:
                result = subprocess.check_call(cmd)
            except Exception as e:
                error = e
        # capture outpuut
        else:
            try:
                result = subprocess.check_output(cmd)
            except Exception as e:
                error = e
            else:
                result = result.decode(sys.stdout.encoding)

            if expected_output is not None:
                error = result != expected_output

        if error:
            lines = []

            if error_message:
                line_first = error_message
            else:
                if show_cmd:
                    cmd_str = ' '.join(cmd)
                    line_first = ('An error occurred when executing '
                                  f'command: {cmd_str}')
                else:
                    line_first = 'An error occurred.'

            lines.append(line_first)

            if not capture_output:
                lines.append(f'Original error message: {error}')

            if hint:
                lines.append(f'Hint: {hint}.')

            raise CommanderException('\n'.join(lines))
        else:
            return result

    def __enter__(self):
        if self.workspace and not Path(self.workspace).exists():
            Path(self.workspace).mkdir()

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # move to the original working directory
        os.chdir(self._wd)

        self.rm(*self._to_delete)
        supress = isinstance(exc_value, CommanderStop)

        if supress:
            self.info(str(exc_value))

        self._warn_show()

        return supress

    def rm(self, *args):
        """Deletes all files/directories

        Examples
        --------
        >>> cmdr.rm('file', 'directory') # doctest: +SKIP
        """
        for f in args:
            _delete(f)

    def rm_on_exit(self, path):
        """Removes file upon exit

        Examples
        --------
        >>> cmdr.rm_on_exit('some_temporary_file') # doctest: +SKIP
        """
        self._to_delete.append(Path(path).resolve())

    def copy_template(self, path, **render_kwargs):
        """Copy template to the workspace

        Parameters
        ----------
        path : str
            Path to template (relative to templates path)

        **render_kwargs
            Keyword arguments passed to the template

        Examples
        --------
        >>> # copies template in {templates-path}/directory/template.yaml
        >>> # to {workspace}/template.yaml
        >>> cmdr.copy_template('directory/template.yaml') # doctest: +SKIP
        """
        dst = Path(self.workspace, PurePosixPath(path).name)

        # This message is no longer valid since this is only called
        # when there is no env yet
        if dst.exists():
            self.success(f'Using existing {path!s}...')
        else:
            self.info(f'Adding {dst!s}...')
            dst.parent.mkdir(exist_ok=True, parents=True)
            content = self._env.get_template(str(path)).render(**render_kwargs)
            dst.write_text(content)

    def cd(self, dir_):
        """Change current working directory
        """
        os.chdir(dir_)

    def cp(self, src):
        """
        Copies a file or directory to the workspace, replacing it if necessary.
        Deleted on exit.

        Notes
        -----
        Used mainly for preparing Dockerfiles since they can only
        copy from the current working directory

        Examples
        --------
        >>> # copies dir/file to {workspace}/file
        >>> cmdr.cp('dir/file') # doctest: +SKIP
        """
        path = Path(src)

        if not path.exists():
            raise CommanderException(
                f'Missing {src} file. Add it and try again.')

        # convert to absolute to ensure we delete the right file on __exit__
        dst = Path(self.workspace, path.name).resolve()
        self._to_delete.append(dst)

        _delete(dst)

        if path.is_file():
            shutil.copy(src, dst)
        else:
            shutil.copytree(src, dst)

    def append_inline(self, line, dst):
        """Append line to a file

        Parameters
        ----------
        line : str
            Line to append

        dst : str
            File to append (can be outside the workspace)

        Examples
        --------
        >>> cmdr.append_inline('*.csv', '.gitignore') # doctest: +SKIP
        """
        if not Path(dst).exists():
            Path(dst).touch()

        original = Path(dst).read_text()
        Path(dst).write_text(original + '\n' + line + '\n')

    def print(self, line):
        """Print message (no color)
        """
        self.tw.write(f'{line}\n')

    def success(self, line=None):
        """Print success message (green)
        """
        self.tw.sep('=', line, green=True)

    def info(self, line=None):
        """Print information message (blue)
        """
        self.tw.sep('=', line, blue=True)

    def warn(self, line=None):
        """Print warning (yellow)
        """
        self.tw.sep('=', line, yellow=True)

    def warn_on_exit(self, line):
        """Append a warning message to be displayed on exit
        """
        self._warnings.append(line)

    def _warn_show(self):
        """Display accumulated warning messages (added via .warn_on_exit)
        """
        if self._warnings:
            self.tw.sep('=', 'Warnings', yellow=True)
            self.tw.write('\n\n'.join(self._warnings) + '\n')
            self.tw.sep('=', yellow=True)



from pathlib import Path


def iterable(obj, repr_=False):
    if repr_:
        sorted_ = sorted(repr(element) for element in obj)
    else:
        sorted_ = sorted(f"'{element}'" for element in obj)

    if len(sorted_) > 1:
        sorted_[-1] = f'and {sorted_[-1]}'

    return ", ".join(sorted_)


def them_or_name(obj):
    return 'them' if len(obj) > 1 else f"'{list(obj)[0]}'"


def trailing_dot(obj):
    return ' '.join(f'{e}.' for e in obj)


def try_relative_path(path):
    if not isinstance(path, (str, Path)):
        return path

    path_ = Path(path)

    if path_.is_absolute():
        try:
            return str(path_.relative_to(Path().resolve()))
        except ValueError:
            return str(path)
    else:
        return str(path)



import logging


class DAGLogger:
    """Set up logging when calling DAG.build()

    DAGLogger logger works by simply adding a logger handler to the passed
    logger when DAG.build() starts and removing it when it finishes. This class
    is not intended to be used directly.

    """
    def __init__(self, handler, logger=None):
        self.handler = handler

        # if no logger is passed, set it to the root logger with debug level
        if logger is None:
            self.logger = logging.getLogger()
            self.logger.setLevel(logging.INFO)
        else:
            self.logger = logger

    def __enter__(self):
        # can be None
        if self.handler:
            self.logger.addHandler(self.handler)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.handler:
            self.logger.removeHandler(self.handler)



# from: https://github.com/duelafn/python-jinja2-apci
from jinja2 import nodes
from jinja2.ext import Extension
from jinja2.exceptions import TemplateRuntimeError


class RaiseExtension(Extension):
    """
    jinja template extension to allow raising exceptions
    """
    tags = set(['raise'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        message_node = parser.parse_expression()
        return nodes.CallBlock(self.call_method('_raise', [message_node],
                                                lineno=lineno), [], [], [],
                               lineno=lineno)

    def _raise(self, msg, caller):
        raise TemplateRuntimeError(msg)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import re
import ast
from glob import glob
from os.path import basename
from os.path import splitext
from pathlib import Path

from setuptools import find_packages
from setuptools import setup

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('src/ploomber/__init__.py', 'rb') as f:
    VERSION = str(
        ast.literal_eval(
            _version_re.search(f.read().decode('utf-8')).group(1)))

here = Path(__file__).parent.resolve()


def read(name):
    return Path(here, name).read_text(encoding='utf-8')


# NOTE: most users just do "pip install jupyter" but
# we have to pin specific versions of jupyter_client, nbconvert and
# ipykernel to support parallel execution using papermill
# (avoid "kernel did not respond" errors)
# these are versions are not pinned in papermill (yet) so we put it here
# more info:
# https://discourse.jupyter.org/t/nbconvert-5-6-0-release/1867
# https://github.com/nteract/papermill/issues/239
NB = [
    'papermill',
    'jupytext',
    'ipykernel>=1.5.2',
    'jupyter_client>=5.3.1',
    'nbconvert>=5.6.0',
    'notebook',
    'nbformat',
    # for notebook validation
    'pyflakes',
]

# Optional dependencies are packages that are used in several modules but are
# not strictly required. Dependencies that are required for a single use case
# (e.g. upload to s3) should be included in the "TESTING" list. Both optional
# and one-time modules should use the @requires decorator to show an error if
# the dependency is missing. numpydoc is an special case because it's an
# optional dependency but not having it installed does not trigger an error
# it will just not print the parsed docstring.
OPTIONAL = [
    # sql dumps and uploads
    'pandas',
    # for ParquetIO
    'pyarrow',
    # qa and entry modules
    'numpydoc',

    # for embedded dag plots with d3 backend
    'requests-html',
    'nest_asyncio',
]

TESTING = [
    # plotting. strictly speaking pygrapviz is an optional dependency but we
    # don't add it as such because it's gonna break installation for most
    # setups, since we don't expect users to have graphviz installed
    'pygraphviz',
    # RemoteShellClient
    'paramiko',
    # Upload to S3
    'boto3',
    # testing upload to S3 task
    'moto',
    # Upload to google cloud storage
    'google-cloud-storage',
    'pytest',
    'pytest-cov',
    # TODO: update config so coveralls 3 works
    'coveralls<3',
    # we need this because we are re-using the original jupyter test suite for
    # testing our contents manager (which imports nose), see test_jupyter.py
    'nose',
    'yapf',
    'flake8',
    # needed to run some test pipelines
    'matplotlib',
    'seaborn',
    # https://www.psycopg.org/docs/install.html#psycopg-vs-psycopg-binary
    # this one is easier to install
    'psycopg2-binary',
    # requires for some Table tests where we parse the HTML repr
    'lxml',

    # for testing jupyter lab plugin
    'jupyter_server',

    # optional dependencies for @serializer and @unserializer
    'joblib',
    'cloudpickle',

    # for testing the webpdf converter
    'nbconvert[webpdf]',
]

# packages needed for development
DEV = ['twine', 'invoke', 'pkgmt']

DESCRIPTION = (
    'Write maintainable, production-ready pipelines using Jupyter or your '
    'favorite text editor. Develop locally, deploy to the cloud.')

setup(
    name='ploomber',
    version=VERSION,
    description=DESCRIPTION,
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='Ploomber',
    author_email='contact@ploomber.io',
    url='https://github.com/ploomber/ploomber',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    data_files=[
        ("etc/jupyter/jupyter_notebook_config.d",
         ["jupyter-config/jupyter_notebook_config.d/ploomber.json"]),
        ("etc/jupyter/jupyter_server_config.d",
         ["jupyter-config/jupyter_server_config.d/ploomber.json"]),
    ],
    zip_safe=False,
    classifiers=[
        # https://pypi.org/classifiers/
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    install_requires=[
        'ploomber-scaffold>=0.3',
        'pyyaml',
        'networkx>=2.5',
        # nbconvert is broken with jinja>=3.1
        # https://github.com/jupyter/nbconvert/issues/1736
        'jinja2<3.1',
        'tabulate',
        'humanize',
        'tqdm',
        'posthog',
        'importlib_resources;python_version<"3.7"',
        # for code normalization, parso is also needed for inferring upstream
        # dependencies in jupyter notebooks
        'sqlparse',
        'autopep8',
        'parso',
        # for generating dag.to_markup(fmt='html')
        'mistune',
        # for syntax highlighting when generating dag HTML reports
        'pygments',
        'sqlalchemy',
        # for cli
        'click',
        # for ploomber interact and {PythonCallable, NotebookRunner}.debug()
        'ipython<8',
        'ipdb',
        'pydantic',
    ] + NB,
    extras_require={
        'all': OPTIONAL,
        'dev': OPTIONAL + TESTING + DEV,
    },
    entry_points={
        'console_scripts': ['ploomber=ploomber_cli.cli:cmd_router'],
    },
)



from ploomber.validators.validators import (data_frame_validator,
                                            validate_schema, validate_values,
                                            Assert)

__all__ = [
    'data_frame_validator', 'validate_schema', 'validate_values', 'Assert'
]



"""
Task.exec_status only makes sense for parallel execution, when using the serial
executor, we can update the _status automatically (within Task) but it is
unusable, get rid of the logic that updates it automatically, however,
we ar eusing here the _update status here to propagate dowstream updates.
So think about which parts can go away and which ones should not
"""
import os
import logging
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures.process import BrokenProcessPool
from ploomber.constants import TaskStatus
from ploomber.executors.abc import Executor
from ploomber.exceptions import DAGBuildError
from ploomber.messagecollector import (BuildExceptionsCollector, Message)
from ploomber.executors import _format

# TODO: support for show_progress, we can use a progress bar but we have
# to modify the label since at any point more than one task might be
# executing

# TODO: add support for logging errors, then add this executor to the tests
# in test_hooks


class TaskBuildWrapper:
    """
    Wraps task._build. Traceback is lost when using ProcessPoolExecutor
    (future.exception().__traceback__ returns None), so we have to catch it
    here and return it so at the end of the DAG execution, a message with the
    traceback can be printed
    """
    def __init__(self, task):
        self.task = task

    def __call__(self, *args, **kwargs):
        try:
            output = self.task._build(**kwargs)
            return output
        except Exception as e:
            return Message(task=self.task, message=_format.exception(e), obj=e)


def _log(msg, logger, print_progress):
    logger(msg)

    if print_progress:
        print(msg)


class Parallel(Executor):
    """Runs a DAG in parallel using multiprocessing

    Parameters
    ----------
    processes : int, default=None
        The number of processes to use. If None, uses ``os.cpu_count``

    print_progress : bool, default=False
        Whether to print progress to stdout, otherwise just log it

    Examples
    --------
    Spec API:

    .. code-block:: yaml
        :class: text-editor
        :name: pipeline-yaml

        # add at the top of your pipeline.yaml
        executor: parallel


    Python API:

    >>> from ploomber import DAG
    >>> from ploomber.executors import Parallel
    >>> dag = DAG(executor='parallel') # use with default values
    >>> dag = DAG(executor=Parallel(processes=2)) # customize


    Notes
    -----
    If any task crashes, downstream tasks execution is aborted, building
    continues until no more tasks can be executed
    """

    # NOTE: Tasks should not create child processes
    # https://docs.python.org/3/library/multiprocessing.html#multiprocessing.Process.daemon

    def __init__(self, processes=None, print_progress=False):
        self.processes = processes or os.cpu_count()
        self.print_progress = print_progress

        self._logger = logging.getLogger(__name__)
        self._i = 0

    def __call__(self, dag, show_progress):
        super().__call__(dag)

        # TODO: Have to test this with other Tasks, especially the ones that
        # use clients - have to make sure they are serialized correctly
        done = []
        started = []
        set_all = set(dag)
        future_mapping = {}

        # there might be up-to-date tasks, add them to done
        # FIXME: this only happens when the dag is already build and then
        # then try to build again (in the same session), if the session
        # is restarted even up-to-date tasks will be WaitingExecution again
        # this is a bit confusing, so maybe change WaitingExecution
        # to WaitingBuild?
        for name in dag:
            if dag[name].exec_status in {
                    TaskStatus.Executed, TaskStatus.Skipped
            }:
                done.append(dag[name])

        def callback(future):
            """Keep track of finished tasks
            """
            task = future_mapping[future]
            self._logger.debug('Added %s to the list of finished tasks...',
                               task.name)
            try:
                result = future.result()
            except BrokenProcessPool:
                # ignore the error here but flag the task,
                # so next_task is able to stop the iteration,
                # when we call result after breaking the loop,
                # this will show up
                task.exec_status = TaskStatus.BrokenProcessPool
            else:
                if isinstance(result, Message):
                    task.exec_status = TaskStatus.Errored
                # sucessfully run task._build
                else:
                    # ignore report here, we just the metadata to update it
                    _, meta = result
                    task.product.metadata.update_locally(meta)
                    task.exec_status = TaskStatus.Executed

            done.append(task)

        def next_task():
            """
            Return the next Task to execute, returns None if no Tasks are
            available for execution (cause their dependencies are not done yet)
            and raises a StopIteration exception if there are no more tasks to
            run, which means the DAG is done
            """
            for task in dag.values():
                if task.exec_status in {TaskStatus.Aborted}:
                    done.append(task)
                elif task.exec_status == TaskStatus.BrokenProcessPool:
                    raise StopIteration

            # iterate over tasks to find which is ready for execution
            for task in dag.values():
                # ignore tasks that are already started, I should probably add
                # an executing status but that cannot exist in the task itself,
                # maybe in the manaer?
                if (task.exec_status in {
                        TaskStatus.WaitingExecution, TaskStatus.WaitingDownload
                } and task not in started):
                    return task
                # there might be some up-to-date tasks, add them

            set_done = set([t.name for t in done])

            if not self._i % 50000:
                _log(f'Finished tasks so far: {set_done or ""}',
                     self._logger.debug,
                     print_progress=self.print_progress)
                _log(f'Remaining tasks: {set_all - set_done}',
                     self._logger.debug,
                     print_progress=self.print_progress)
                _log(f'Finished {len(set_done)} out of {len(set_all)} tasks',
                     self._logger.info,
                     print_progress=self.print_progress)

            if set_done == set_all:
                self._logger.debug('All tasks done')

                raise StopIteration

            self._i += 1

        task_kwargs = {'catch_exceptions': True}

        with ProcessPoolExecutor(max_workers=self.processes) as pool:
            while True:
                try:
                    task = next_task()
                except StopIteration:
                    break
                else:
                    if task is not None:
                        future = pool.submit(TaskBuildWrapper(task),
                                             **task_kwargs)
                        # the callback function uses the future mapping
                        # so add it before registering the callback, otherwise
                        # it might break and hang the whole process
                        future_mapping[future] = task
                        future.add_done_callback(callback)
                        started.append(task)
                        self._logger.info('Added %s to the pool...', task.name)

        results = [
            # results are the output of Task._build: (report, metadata)
            # OR a Message
            get_future_result(f, future_mapping)
            for f in future_mapping.keys()
        ]

        exps = [r for r in results if isinstance(r, Message)]

        if exps:
            raise DAGBuildError(str(BuildExceptionsCollector(exps)))

        # if we reach this, it means no tasks failed. only return reports
        return [r[0] for r in results]

    def __getstate__(self):
        state = self.__dict__.copy()
        # _logger is not pickable, so we remove them and build
        # them again in __setstate__
        del state['_logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._logger = logging.getLogger(__name__)


def get_future_result(future, future_mapping):
    try:
        return future.result()
    except BrokenProcessPool as e:
        raise BrokenProcessPool('Broken pool {}'.format(
            future_mapping[future])) from e



"""
Task implementations

A Task is a unit of work that produces a persistent change (Product)
such as a bash or a SQL script
"""
import shlex
import subprocess
import pdb
import functools
from collections.abc import Mapping

from IPython.terminal.debugger import TerminalPdb, Pdb

from ploomber.tasks.abc import Task
from ploomber.tasks.mixins import ClientMixin
from ploomber.sources import (PythonCallableSource, GenericSource, EmptySource)
from ploomber.clients import ShellClient
from ploomber.products.metadata import MetadataAlwaysUpToDate
from ploomber.exceptions import TaskBuildError, MissingClientError
from ploomber.constants import TaskStatus
from ploomber.sources.interact import CallableInteractiveDeveloper
from ploomber.tasks._params import Params
from ploomber.io.loaders import _file_load
from ploomber.products import MetaProduct


def _unserializer(product, unserializer):
    # this happens when we have a task group. e.g., product declares
    # 'fit-*' as dependency, which matches fit-1, fit-2, etc. Then product
    # becomes...
    # {'fit-*': {'fit-1': product, 'fit-2': another, ...}}
    if isinstance(product, Mapping) and not isinstance(product, MetaProduct):
        return {k: unserializer(p) for k, p in product.items()}
    else:
        # pass metaproduct and products directly to the unserializer function
        return unserializer(product)


def _unserialize_params(params_original, unserializer):
    """
    User the user-provided function to unserialize params['upstream']
    """
    params = params_original.to_dict()

    params['upstream'] = {
        k: _unserializer(v, unserializer)
        for k, v in params['upstream'].items()
    }

    params = Params._from_dict(params, copy=False)

    return params


class PythonCallable(Task):
    """
    Run a Python callable (e.g. a function)

    Parameters
    ----------
    source: callable
        The callable to execute
    product: ploomber.products.Product
        Product generated upon successful execution
    dag: ploomber.DAG
        A DAG to add this task to
    name: str
        A str to indentify this task. Should not already exist in the dag
    params: dict
        Parameters to pass to the callable, by default, the callable will
        be executed with a "product" (which will contain the product object).
        It will also include a "upstream" parameter if the task has upstream
        dependencies along with any parameters declared here
    unserializer: callable, optional
        A callable to unserialize upstream products, the product object
        is passed as unique argument. If None, the source function receives
        the product object directly. If the task has no upstream dependencies,
        this argument has no effect
    serializer: callable, optional
        A callable to serialize this task's product, must take two arguments,
        the first argument passed is the value returned by the task's
        source, the second argument is the product oject. If None, the
        task's source is responsible for serializing its own product. If
        used, the source function must not have a "product" parameter but
        return its result instead
    """
    def __init__(self,
                 source,
                 product,
                 dag,
                 name=None,
                 params=None,
                 unserializer=None,
                 serializer=None):
        self._serializer = serializer or dag.serializer
        kwargs = dict(hot_reload=dag._params.hot_reload,
                      needs_product=self._serializer is None)
        self._source = type(self)._init_source(source, kwargs)
        self._unserializer = unserializer or dag.unserializer
        super().__init__(product, dag, name, params)

    @staticmethod
    def _init_source(source, kwargs):
        return PythonCallableSource(source, **kwargs)

    def run(self):
        if 'upstream' in self.params and self._unserializer:
            params = _unserialize_params(self.params, self._unserializer)
        else:
            params = self.params.to_dict()

        # do not pass product if serializer is set, we'll use the returned
        # value in such case
        if self._serializer:
            product = params.pop('product')
        else:
            product = params['product']

        # call function
        out = self.source.primitive(**params)

        # serialize output if needed
        if self._serializer:
            if out is None:
                raise ValueError('Callable {} must return a value if task '
                                 'is initialized with a serializer'.format(
                                     self.source.primitive))
            else:
                self._serializer(out, product)

    def develop(self, app='notebook', args=None):
        """Edit function interactively using Jupyter

        Parameters
        ----------
        app : str, {'notebook', 'lab'}, default='notebook'
            Which jupyter application to use

        args : str
            Extra args passed to the selected jupyter application

        Notes
        -----
        Cells whose first line is an empty commenf ("#"), will be removed when
        exporting back to the function, you can use this for temporary,
        exploratory work

        Be careful when developing tasks interacively. If the task has run
        successfully, you overwrite products but don't save the
        updated source code, your DAG will enter an inconsistent state where
        the metadata won't match the overwritten product.
        """
        apps = {'notebook', 'lab'}

        if app not in apps:
            raise ValueError('"app" must be one of {}, got: "{}"'.format(
                apps, app))

        if self.exec_status == TaskStatus.WaitingRender:
            raise TaskBuildError(
                'Error in task "{}". '
                'Cannot call task.develop() on a task that has '
                'not been '
                'rendered, call DAG.render() first'.format(self.name))

        with self._interactive_developer() as tmp:
            try:
                subprocess.run(['jupyter', app, tmp] + shlex.split(args or ''),
                               check=True)
            except KeyboardInterrupt:
                print(f'Jupyter {app} application closed...')

    def _interactive_developer(self):
        """
        Creates a CallableInteractiveDeveloper instance
        """
        # TODO: resolve to absolute to make relative paths work
        return CallableInteractiveDeveloper(self.source.primitive, self.params)

    def debug(self, kind='ipdb'):
        """
        Run callable in debug mode.

        Parameters
        ----------
        kind : str ('ipdb' or 'pdb')
            Which debugger to use 'ipdb' for IPython debugger or 'pdb' for
            debugger from the standard library

        Notes
        -----
        Be careful when debugging tasks. If the task has run
        successfully, you overwrite products but don't save the
        updated source code, your DAG will enter an inconsistent state where
        the metadata won't match the overwritten product.
        """
        opts = {'ipdb', 'pdb'}

        if kind not in opts:
            raise ValueError('"kind" must be one of {}, got: "{}"'.format(
                opts, kind))

        if self.exec_status == TaskStatus.WaitingRender:
            raise TaskBuildError('Error in task "{}". '
                                 'Cannot call task.debug() on a task that has '
                                 'not been '
                                 'rendered, call DAG.render() first'.format(
                                     self.name))

        if 'upstream' in self.params and self._unserializer:
            params = _unserialize_params(self.params, self._unserializer)
        else:
            params = self.params.to_dict()

        if self._serializer:
            params.pop('product')

        if kind == 'ipdb':
            try:
                # this seems to only work in a Terminal
                ipdb = TerminalPdb()
            except Exception:
                # this works in a Jupyter notebook
                ipdb = Pdb()

            ipdb.runcall(self.source.primitive, **params)
        elif kind == 'pdb':
            pdb.runcall(self.source.primitive, **params)

    def load(self, key=None, **kwargs):
        """
        Loads the product. It uses the unserializer function if any, otherwise
        it tries to load it based on the file extension

        Parameters
        ----------
        key
            Key to load, if this task generates more than one product

        **kwargs
            Arguments passed to the unserializer function
        """
        if isinstance(self.product, MetaProduct) and key is None:
            raise ValueError(f'Task {self!r} generates multiple products, '
                             'use the "key" argument to load one')

        prod = self.product if not key else self.product[key]

        if self._unserializer is not None:
            return self._unserializer(str(prod), **kwargs)
        else:
            return _file_load(prod, **kwargs)


# FIXME: there is already a TaskFactory, this is confusing
def task_factory(_func=None, **factory_kwargs):
    """Syntactic sugar for building PythonCallable tasks
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(**wrapper_kwargs):
            kwargs = {**factory_kwargs, **wrapper_kwargs}
            return PythonCallable(func, **kwargs)

        return wrapper

    return decorator if _func is None else decorator(_func)


class ShellScript(ClientMixin, Task):
    """Execute a shell script in a shell

    Parameters
    ----------
    source: str or pathlib.Path
        Script source, if str, the content is interpreted as the actual
        script, if pathlib.Path, the content of the file is loaded. The
        souce code must have the {{product}} tag
    product: ploomber.products.Product
        Product generated upon successful execution
    dag: ploomber.DAG
        A DAG to add this task to
    name: str
        A str to indentify this task. Should not already exist in the dag
    client: ploomber.clients.ShellClient or RemoteShellClient, optional
        The client used to connect to the database. Only required
        if no dag-level client has been declared using dag.clients[class]
    params: dict, optional
        Parameters to pass to the script, by default, the callable will
        be executed with a "product" (which will contain the product object).
        It will also include a "upstream" parameter if the task has upstream
        dependencies along with any parameters declared here. The source
        code is converted to a jinja2.Template for passing parameters,
        refer to jinja2 documentation for details
    """
    def __init__(self,
                 source,
                 product,
                 dag,
                 name=None,
                 client=None,
                 params=None):
        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = type(self)._init_source(source, kwargs)
        super().__init__(product, dag, name, params)
        self._client = client

    @property
    def client(self):
        try:
            client = super().client
        except MissingClientError:
            self._client = ShellClient()
            return self._client
        else:
            return client

    @staticmethod
    def _init_source(source, kwargs):
        required = {
            'product': ('ShellScript must include {{product}} in '
                        'its source')
        }

        return GenericSource(source, **kwargs, required=required)

    def run(self):
        self.client.execute(str(self.source))


class DownloadFromURL(Task):
    """
    Download a file from a URL (uses urllib.request.urlretrieve)

    Parameters
    ----------
    source: str
        URL to download the file from
    product: ploomber.products.File
        Product generated upon successful execution
    dag: ploomber.DAG
        A DAG to add this task to
    name: str
        A str to indentify this task. Should not already exist in the dag
    """
    def __init__(self, source, product, dag, name=None, params=None):
        params = params or {}
        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = type(self)._init_source(source, kwargs)
        super().__init__(product, dag, name, params)

    def run(self):
        # lazily load urllib as it is slow to import
        from urllib import request
        request.urlretrieve(str(self.source), filename=str(self.product))

    @staticmethod
    def _init_source(source, kwargs):
        return GenericSource(str(source), **kwargs, optional=['product'])


class Link(Task):
    """
    A dummy Task used to "plug" an external Product to a pipeline, this
    task is always considered up-to-date

    The purpose of this Task is to link a pipeline to an external read-only
    file, this task does not do anything on the dataset and the product is
    always considered up-to-date. There are two primary use cases:
    when the raw data is automatically uploaded to a file (or table) and the
    pipeline does not have control over data updates, this task can be used
    to link the pipeline to that file, without having to copy it, downstream
    tasks will see this dataset as just another Product. The second use case
    is when developing a prediction pipeline. When making predictions on new
    data, the pipeline might rely on existing data to generate features,
    this task can be used to point to such file it can also be used to
    point to a serialized model, this last scenario is only recommended
    for prediction pipeline that do not have strict performance requirements,
    unserializing models is an expensive operation, for real-time predictions,
    the model should be kept in memory

    Parameters
    ----------
    product: ploomber.products.Product
        Product to link to the dag
    dag: ploomber.DAG
        A DAG to add this task to
    name: str
        A str to indentify this task. Should not already exist in the dag
    """
    def __init__(self, product, dag, name):
        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = type(self)._init_source(kwargs)
        super().__init__(product, dag, name, None)

        # patch product's metadata
        self.product.metadata = MetadataAlwaysUpToDate()
        # product's code will never be considered outdated
        self.product._outdated_code_dependency = self._false

        if not self.product.exists():
            raise RuntimeError('Link tasks should point to Products that '
                               'already exist. "{}" task product "{}" does '
                               'not exist'.format(self.name, self.product))

    def run(self):
        pass

    def set_upstream(self, other):
        raise RuntimeError('Link tasks should not have upstream dependencies')

    @staticmethod
    def _init_source(kwargs):
        return EmptySource(None, **kwargs)

    def _false(self):
        # this should be __false but we can't due to
        # https://bugs.python.org/issue33007
        return False


class Input(Task):
    """
    A dummy task used to represent input provided by the user, it is always
    considered outdated.

    When making new predictions, the user must submit some input data to build
    features and then feed the model, this task can be used to point to
    such input. It does not perform any processing (read-only data) but it is
    always considered outdated, which means it will always trigger execution.

    Parameters
    ----------
    product: ploomber.products.Product
        Product to to serve as input to the dag
    dag: ploomber.DAG
        A DAG to add this task to
    name: str
        A str to indentify this task. Should not already exist in the dag
    """
    def __init__(self, product, dag, name):
        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = type(self)._init_source(kwargs)
        super().__init__(product, dag, name, None)

        # do not save metadata (Product's location is read-only)
        self.product.metadata.update = self._null_update_metadata

        # the product will always be considered outdated
        self.product._outdated_data_dependencies = self._true
        self.product._outdated_code_dependency = self._true

        if not self.product.exists():
            raise RuntimeError('Input tasks should point to Products that '
                               'already exist. "{}" task product "{}" does '
                               'not exist'.format(self.name, self.product))

    def run(self):
        pass

    def set_upstream(self, other):
        raise RuntimeError('Input tasks should not have upstream '
                           'dependencies')

    @staticmethod
    def _init_source(kwargs):
        return EmptySource(None, **kwargs)

    def _null_update_metadata(self, source_code, params):
        # this should be __null_update_metadata but we can't due to
        # https://bugs.python.org/issue33007
        pass

    def _true(self):
        # this should be __true but we can't due to
        # https://bugs.python.org/issue33007
        return True



from ploomber.cli.parsers import CustomParser
from ploomber.cli.io import cli_endpoint
from ploomber.telemetry import telemetry


@cli_endpoint
@telemetry.log_call('report')
def main():
    parser = CustomParser(description='Make a pipeline report',
                          prog='ploomber report')
    with parser:
        parser.add_argument(
            '--output',
            '-o',
            help='Where to save the report, defaults to pipeline.html',
            default='pipeline.html')

    dag, args = parser.load_from_entry_point_arg()
    dag.to_markup(path=args.output)

    print('Report saved at:', args.output)



from copy import copy
import warnings
from collections import defaultdict, abc
from collections.abc import Mapping

from ploomber.products import MetaProduct
from ploomber.exceptions import UpstreamKeyError


class Upstream(abc.Mapping):
    """
    Mapping for representing upstream tasks, it's a collections.OrderedDict
    under the hood with an added .pop(key) method (like the one in a regular
    dictionary) and with a .first attribute, that returns the first value,
    useful when there is only one key-value pair.

    Used for task upstream dependencies, when employed as a context manager,
    it raises a warning if any of the elements was not used at least once,
    which means there are unused upstream dependencies

    Parameters
    ----------
    data : dict
        Data with upstream dependencies

    name : str
        Task name, only used to give more context when raising an exception
        (e.g. trying to access a key that does not exist)
    """
    def __init__(self, data=None, name=None):
        self._dict = (data or {})
        self._init_counts()
        self._in_context = False
        self._name = name

    # we cannot define this inside _init_counts or as a lambda, that does
    # not work with pickle
    def _zero():
        return 0

    def _init_counts(self):
        # lambdas do not work with pickle, must define a function
        self._counts = defaultdict(self._zero,
                                   {key: 0
                                    for key in self._dict.keys()})

    @property
    def first(self):
        if not self._dict:
            raise KeyError('Cannot obtain first upstream task, task "{}" has '
                           'no upstream dependencies declared'.format(
                               self._name))
        if len(self._dict) > 1:
            raise ValueError('first can only be used if there is a single '
                             'upstream dependency, task "{}" has: {}'.format(
                                 self._name, len(self._dict)))

        first_key = next(iter(self._dict))
        return self._dict[first_key]

    def pop(self, key):
        return self._dict.pop(key)

    # NOTE: are we using this?
    def to_dict(self):
        return copy(self._dict)

    def to_json_serializable(self):
        """
        Converts the object to a dictionary, calls .to_json_serializable
        on all products. In case of grouped products (e.g., when using
        group_name in task.set_upstream), it calls .to_json_serializable
        on each element of the group
        """
        return {k: _to_json_serializable(v) for k, v in self._dict.items()}

    def __getitem__(self, key):
        if self._in_context:
            self._counts[key] += 1

        if not len(self._dict):
            raise UpstreamKeyError(
                'Cannot obtain upstream dependency "{}". '
                'Task "{}" has no upstream dependencies'.format(
                    key, self._name))

        try:
            return self._dict[key]
        except KeyError:
            raise UpstreamKeyError(
                'Cannot obtain upstream dependency "{}" for task "{}" '
                'declared dependencies are: {}'.format(key, self._name, self))

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __iter__(self):
        for name in self._dict.keys():
            yield name

    def __len__(self):
        return len(self._dict)

    def __enter__(self):
        self._in_context = True
        self._init_counts()
        return self

    def __str__(self):
        return str(self._dict)

    def __repr__(self):
        return 'Upstream({})'.format(repr(self._dict))

    def __exit__(self, *exc):
        self._in_context = False
        unused = set(
            [key for key, count in self._counts.items() if count == 0])

        if unused:
            warnings.warn('The following upstream dependencies in task "{}" '
                          'were not used {}'.format(self._name, unused))


def _to_json_serializable(value):
    # when value is a mapping but not a metaproduct, we are dealing with
    # a task group and need to process elements individually
    if isinstance(value, Mapping) and not isinstance(value, MetaProduct):
        return {k: v.to_json_serializable() for k, v in value.items()}
    else:
        return value.to_json_serializable()



import sys
import uuid

from ploomber.cli.parsers import CustomParser
from ploomber.cli.io import cli_endpoint
from ploomber.executors import Parallel
from ploomber.telemetry import telemetry
from ploomber.cli.cloud import _write_pipeline


# this parameter is only set to True when calling "ploomber interactive"
@cli_endpoint
@telemetry.log_call('build', payload=True)
def main(payload, render_only=False):
    parser = CustomParser(description='Build pipeline', prog='ploomber build')

    with parser:
        parser.add_argument('--force',
                            '-f',
                            help='Force execution by ignoring status',
                            action='store_true',
                            default=False)
        parser.add_argument('--skip-upstream',
                            '-su',
                            help='Skip building upstream dependencies. '
                            'Only applicable when using --partially',
                            action='store_true',
                            default=False)
        parser.add_argument(
            '--partially',
            '-p',
            help='Build a pipeline partially until certain task',
            default=None)
        parser.add_argument(
            '--debug',
            '-d',
            help='Drop a debugger session if an exception happens',
            action='store_true',
            default=False)

    # users may try to run "ploomber build {name}" to build a single task
    if len(sys.argv) > 1 and not sys.argv[1].startswith('-'):
        suggestion = 'ploomber task {task-name}'
        cmd_name = parser.prog
        telemetry.log_api("unsupported_build_cmd",
                          metadata={
                              'cmd_name': cmd_name,
                              'suggestion': suggestion,
                              'argv': sys.argv
                          })
        parser.error(f'{cmd_name!r} does not take positional arguments.\n'
                     f'To build a single task, try: {suggestion!r}')

    dag, args = parser.load_from_entry_point_arg()
    pid = str(uuid.uuid4())
    _write_pipeline(pipeline_id=pid, status='started', pipeline_name=dag.name)

    # when using the parallel executor from the CLI, ensure we print progress
    # to stdout
    if isinstance(dag.executor, Parallel):
        dag.executor.print_progress = True
    try:
        if render_only:
            dag.render()
        else:
            if args.partially:
                report = dag.build_partially(args.partially,
                                             force=args.force,
                                             debug=args.debug,
                                             skip_upstream=args.skip_upstream)
            else:
                report = dag.build(force=args.force, debug=args.debug)
    except Exception as e:
        _write_pipeline(pipeline_id=pid,
                        status='error',
                        pipeline_name=dag.name,
                        log=str(e.args))

        raise

    if report:
        print(report)

    payload['dag'] = dag
    _write_pipeline(pipeline_id=pid,
                    status='finished',
                    pipeline_name=dag.name,
                    dag=telemetry.parse_dag(dag))

    return dag



import json
from pathlib import PurePosixPath, Path

try:
    import boto3
except ImportError:
    boto3 = None

try:
    import botocore
except ImportError:
    botocore = None

from ploomber.util.default import find_root_recursively
from ploomber.util.util import requires
from ploomber.clients.storage.abc import AbstractStorageClient
from ploomber.exceptions import RemoteFileNotFound, DAGSpecInvalidError


class S3Client(AbstractStorageClient):
    """Client for uploading File products to Amazon S3

    Parameters
    ----------
    bucket_name : str
        Bucket to use

    parent : str
        Parent folder in the bucket to store files. For example, if
        ``parent='path/to'``, and a product in your pipeline is
        ``out/data.csv``, your file will appea in the bucket at
        ``path/to/out/data.csv``.

    json_credentials_path : str, default=None
        JSON file to authenticate the client. Must contain
        ``aws_access_key_id`` and ``aws_secret_access_key``. If None, client
        is initialized without arguments (i.e., ``boto3.client('s3')``)

    path_to_project_root : str, default=None
        Path to project root. If None, looks it
        up automatically and assigns it to the parent folder of your
        ``pipeline.yaml`` spec or ``setup.py`` (if your project is a package).
        This determines the path in remote storage. For example, if
        ``path_to_project_root`` is ``/my-project``, you're storing a
        product at ``/my-project/out/data.csv``, and ``parent='some-dir'``,
        the file will be stored in the bucket at ``some-dir/out/data.csv``
        (we first compute the path of your product relative to the project
        root, then prefix it with ``parent``).

    credentials_relative_to_project_root : bool, default=True
        If ``True``, relative paths in ``json_credentials_path`` are so to the
        ``path_to_project_root``, instead of the current working directory

    **kwargs
        Keyword arguments for the client constructor
    """

    @requires(['boto3', 'botocore'], name='S3Client')
    def __init__(self,
                 bucket_name,
                 parent,
                 json_credentials_path=None,
                 path_to_project_root=None,
                 credentials_relative_to_project_root=True,
                 **kwargs):

        if path_to_project_root:
            project_root = path_to_project_root
        else:
            try:
                project_root = find_root_recursively()
            except Exception as e:
                raise DAGSpecInvalidError(
                    f'Cannot initialize {type(self).__name__} because there '
                    'is not project root. Set one or explicitly pass '
                    'a value in the path_to_project_root argument') from e

        self._path_to_project_root = Path(project_root).resolve()

        if (credentials_relative_to_project_root and json_credentials_path
                and not Path(json_credentials_path).is_absolute()):
            json_credentials_path = Path(self._path_to_project_root,
                                         json_credentials_path)

        self._client_kwargs = kwargs

        if json_credentials_path:
            c = json.loads(Path(json_credentials_path).read_text())

            self._client_kwargs = {
                'aws_access_key_id': c['aws_access_key_id'],
                'aws_secret_access_key': c['aws_secret_access_key'],
                **kwargs
            }

        self._client = self._init_client()
        self._parent = parent
        self._bucket_name = bucket_name

    def _init_client(self):
        return boto3.client('s3', **self._client_kwargs)

    def download(self, local, destination=None):
        remote = self._remote_path(local)
        destination = destination or local

        # FIXME: call _download directly and catch the exception to avoid
        # doing to api calls
        if self._is_file(remote):
            self._download(destination, remote)
        else:
            paginator = self._client.get_paginator('list_objects_v2')

            for page in paginator.paginate(Bucket=self._bucket_name,
                                           Prefix=remote):
                if 'Contents' not in page:
                    raise RemoteFileNotFound('Could not download '
                                             f'{local!r} using client {self}: '
                                             'No such file or directory')

                for remote_file in page['Contents']:
                    remote_path = remote_file['Key']
                    rel = PurePosixPath(remote_path).relative_to(remote)
                    destination_file = Path(destination, *rel.parts)
                    destination_file.parent.mkdir(exist_ok=True, parents=True)
                    self._download(str(destination_file), remote_path)

    def _upload(self, local):
        remote = self._remote_path(local)
        self._client.upload_file(str(local), self._bucket_name, remote)

    def _download(self, local, remote):
        Path(local).parent.mkdir(exist_ok=True, parents=True)
        self._client.download_file(self._bucket_name, remote, str(local))

    def _is_file(self, remote):
        resource = boto3.resource('s3', **self._client_kwargs)

        try:
            resource.Object(self._bucket_name, remote).load()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == '404':
                return False
            else:
                raise
        else:
            return True

    def _is_dir(self, remote):
        bucket = boto3.resource('s3', **self._client_kwargs).Bucket(
            self._bucket_name)
        return any(bucket.objects.filter(Prefix=remote))

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_client']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._client = self._init_client()

    def __repr__(self):
        return (f'{type(self).__name__}(bucket_name={self._bucket_name!r}, '
                f'parent={self._parent!r}, '
                f'path_to_project_root={str(self._path_to_project_root)!r})')



"""
Module for the jupyter extension


For debugging:
>>> from ploomber.jupyter.dag import JupyterDAGManager
>>> from ploomber.jupyter.manager import derive_class
>>> from jupytext.contentsmanager import TextFileContentsManager
>>> PloomberContentsManager = derive_class(TextFileContentsManager)
>>> cm = PloomberContentsManager()
"""
import sys
import datetime
import os
import contextlib
from pathlib import Path
from collections.abc import Mapping
from collections import defaultdict

from ploomber.sources.notebooksource import (_cleanup_rendered_nb, inject_cell)
from ploomber.jupyter.dag import JupyterDAGManager
from ploomber.util import loader
from ploomber.exceptions import DAGSpecInvalidError


class DAGMapping(Mapping):
    """dict-like object that maps source files to Tasks
    """
    def __init__(self, pairs):
        default = defaultdict(lambda: [])

        for name, task in pairs:
            default[name].append(task)

        self._mapping = dict(default)

    def __iter__(self):
        for k in self._mapping:
            yield k

    def __getitem__(self, key):
        """Return the first task associated with a given key
        """
        return self._mapping[key][0]

    def __len__(self):
        return len(self._mapping)

    def delete_metadata(self, key):
        """Delete metadata for all tasks associated with a given key
        """
        for task in self._mapping[key]:
            task.product.metadata.delete()

    def __repr__(self):
        return f'{type(self).__name__}({self._mapping}!r)'


@contextlib.contextmanager
def chdir(directory):
    old_dir = os.getcwd()
    try:
        os.chdir(str(directory))
        yield
    finally:
        os.chdir(old_dir)


def resolve_path(parent, path):
    """
    Functions functions resolves paths to make the {source} -> {task} mapping
    work even then `jupyter notebook` is initialized from a subdirectory
    of pipeline.yaml
    """
    try:
        # FIXME: remove :linenumber
        return Path(path).resolve().relative_to(parent).as_posix().strip()
    except ValueError:
        return None


def check_metadata_filter(log, model):
    try:
        cell_metadata_filter = (
            model['content']['metadata']['jupytext']['cell_metadata_filter'])
    except Exception:
        cell_metadata_filter = None

    if cell_metadata_filter == '-all':
        log.warning('Your notebook has filter that strips out '
                    'cell metadata when saving it from the Jupyter notebook '
                    'app. This will prevent you from tagging your '
                    '"parameters" cell. It is possible that this comes '
                    'from jupytext defaults, either add the tag by '
                    'editing the notebook in a text editor or enable '
                    'metadata in the Jupyter app: File -> Jupytext -> '
                    'Include metadata')


def derive_class(base_class):
    # FIXME: is it possible that the base class is jupyter's native contents
    # manager and we need to include jupytext as well?
    class PloomberContentsManager(base_class):
        """
        Ploomber content manager subclasses jupytext TextFileContentsManager
        to keep jupytext features of opening .py files as notebooks but adds
        a feature that automatically injects parameters in notebooks if they
        are part of a pipeline defined in pipeline.yaml, these injected
        parameters are deleted before saving the file
        """
        restart_msg = (' Fix the issue and and restart "jupyter notebook"')

        # TODO: we can cache this depending on the folder where it's called
        # all files in the same folder share the same dag
        def load_dag(self, starting_dir=None, log=True):
            if self.dag is None or self.spec['meta']['jupyter_hot_reload']:
                msg = (
                    '[Ploomber] An error occured when trying to initialize '
                    'the pipeline. Cells won\' be injected until your '
                    'pipeline processes correctly. See error details below.')

                if self.spec and not self.spec['meta']['jupyter_hot_reload']:
                    msg += self.restart_msg

                try:
                    hot_reload = (self.spec
                                  and self.spec['meta']['jupyter_hot_reload'])
                    (self.spec, self.dag,
                     path) = loader.lazily_load_entry_point(
                         starting_dir=starting_dir, reload=hot_reload)
                    path = str(Path(path).resolve())
                # this error means we couldn't locate a parent root (which is
                # required to determine which spec to use). Since it's expected
                # that this happens for some folder, we simply emit a warning
                except DAGSpecInvalidError as e:
                    self.reset_dag()
                    if log:
                        self.log.warning(
                            '[Ploomber] Skipping DAG initialization since '
                            'there isn\'t a project root in the current or '
                            'parent directories. Error message: '
                            f'{str(e)}')
                except Exception:
                    self.reset_dag()
                    if log:
                        self.log.exception(msg)
                else:
                    # dag initialized successfully...
                    # TODO: we have to remove this because the same
                    # jupyter session may more than one pipeline. should
                    # only be in sys path during pipeline loading
                    # NOTE: we only need this when we are using
                    # functions as notebooks
                    base_path = Path(path).resolve()

                    if (self.spec['meta']['jupyter_hot_reload']
                            and base_path not in sys.path):
                        # jupyter does not add the current working dir by
                        # default, if using hot reload and the dag loads
                        # functions from local files, importlib.reload will
                        # fail
                        # NOTE: might be better to only add this when the dag
                        # is actually loading from local files but that means
                        # we have to run some logic and increases load_dag
                        # running time, which we need to be fast
                        sys.path.insert(0, str(base_path))

                    render_success = True

                    with chdir(base_path):
                        # this dag object won't be executed, forcing speeds
                        # up rendering
                        try:
                            self.dag.render(force=True, show_progress=False)
                        except Exception:
                            render_success = False
                            self.reset_dag()
                            if log:
                                msg = ("[Ploomber] An error ocurred when "
                                       "rendering your DAG, cells won't be "
                                       "injected until the issue is resolved")
                                self.log.exception(msg)

                    # if rendering failed, do not continue
                    if not render_success:
                        return

                    if self.spec['meta']['jupyter_functions_as_notebooks']:
                        self.manager = JupyterDAGManager(self.dag)
                    else:
                        self.manager = None

                    pairs = [(resolve_path(
                        Path(self.root_dir).resolve(), t.source.loc), t)
                             for t in self.dag.values()
                             if t.source.loc is not None]
                    dag_mapping = DAGMapping(pairs)

                    did_keys_change = (
                        self.dag_mapping is None
                        or set(self.dag_mapping) != set(dag_mapping))

                    if did_keys_change or self.path != path:
                        self.log.info('[Ploomber] Using dag defined '
                                      f'at: {str(base_path)!r}')
                        self.log.info('[Ploomber] Pipeline mapping keys: '
                                      f'{list(dag_mapping)}')

                    self.dag_mapping = dag_mapping
                    self.path = path

        def reset_dag(self):
            self.spec = None
            self.dag = None
            self.path = None
            self.dag_mapping = None
            self.manager = None

        def __init__(self, *args, **kwargs):
            """
            Initialize the content manger, look for a pipeline.yaml file in the
            current directory, if there is one, load it, if there isn't one
            don't do anything
            """
            # NOTE: there is some strange behavior in Jupyter contents
            # manager. We should not access attributes here (e.g.,
            # self.root_dir) otherwise they aren't correctly initialized
            self.reset_dag()
            super().__init__(*args, **kwargs)

        def get(self, path, content=True, type=None, format=None):
            """
            This is called when a file/directory is requested (even in the list
            view)
            """
            # FIXME: we only need this call when functions as notebooks is true
            # FIXME: reloading inside a (functions) folder causes 404
            if content:
                # this load_dag() is required to list the folder that contains
                # the notebooks exported from functions, however, since jupyter
                # continuosly calls this for the current directory it gets
                # too verbose so we skip showing the log message
                self.load_dag(log=False)

            if self.manager and path in self.manager:
                return self.manager.get(path, content)

            model = super().get(path=path,
                                content=content,
                                type=type,
                                format=format)

            # user requested directory listing, check if there are task
            # functions defined here
            if model['type'] == 'directory' and self.manager:
                if model['content']:
                    model['content'].extend(self.manager.get_by_parent(path))

            check_metadata_filter(self.log, model)

            # if opening a file (ignore file listing), load dag again
            if (model['content'] and model['type'] == 'notebook'):
                # instantiate the dag using starting at the current folder
                self.log.info(
                    f'[Ploomber] Requested model: {model["path"]}. '
                    f'Looking for DAG with root dir: {self.root_dir}')
                self.load_dag(
                    starting_dir=Path(self.root_dir, model['path']).parent)

                if self._model_in_dag(model):
                    self.log.info('[Ploomber] Injecting cell...')
                    inject_cell(model=model,
                                params=self.dag_mapping[model['path']]._params)

            return model

        def save(self, model, path=""):
            """
            This is called when a file is saved
            """
            if self.manager and path in self.manager:
                out = self.manager.overwrite(model, path)
                return out
            else:
                check_metadata_filter(self.log, model)
                # not sure what's the difference between model['path'] and path
                # but path has leading "/", _model_in_dag strips it
                key = self._model_in_dag(model, path)

                if key:
                    content = model['content']
                    metadata = content.get('metadata', {}).get('ploomber', {})

                    if not metadata.get('injected_manually'):
                        self.log.info(
                            '[Ploomber] Cleaning up injected cell in {}...'.
                            format(model.get('name') or ''))
                        model['content'] = _cleanup_rendered_nb(content)

                    self.log.info("[Ploomber] Deleting product's metadata...")
                    self.dag_mapping.delete_metadata(key)

                return super().save(model, path)

        def _model_in_dag(self, model, path=None):
            """Determine if the model is part of the  pipeline
            """
            model_in_dag = False

            if path is None:
                path = model['path']
            else:
                path = path.strip('/')

            if self.dag:
                if ('content' in model and model['type'] == 'notebook'):
                    if path in self.dag_mapping:
                        # NOTE: not sure why sometimes the model comes with a
                        # name and sometimes it doesn't
                        self.log.info(
                            '[Ploomber] {} is part of the pipeline... '.format(
                                model.get('name') or ''))
                        model_in_dag = True
                    else:
                        self.log.info(
                            '[Ploomber] {} is not part of the pipeline, '
                            'skipping...'.format(model.get('name') or ''))

            return path if model_in_dag else False

        def list_checkpoints(self, path):
            if not self.manager or path not in self.manager:
                return self.checkpoints.list_checkpoints(path)

        def create_checkpoint(self, path):
            if not self.manager or path not in self.manager:
                return self.checkpoints.create_checkpoint(self, path)
            else:
                return {
                    'id': 'checkpoint',
                    'last_modified': datetime.datetime.now()
                }

    return PloomberContentsManager


def _load_jupyter_server_extension(app):
    """
    This function is called to configure the new content manager, there are a
    lot of quirks that jupytext maintainers had to solve to make it work so
    we base our implementation on theirs:
    https://github.com/mwouts/jupytext/blob/bc1b15935e096c280b6630f45e65c331f04f7d9c/jupytext/__init__.py#L19
    """
    if hasattr(app.contents_manager_class, 'load_dag'):
        app.log.info("[Ploomber] NotebookApp.contents_manager_class "
                     "is a subclass of PloomberContentsManager already - OK")
        return

    # The server extension call is too late!
    # The contents manager was set at NotebookApp.init_configurables

    # Let's change the contents manager class
    app.log.info('[Ploomber] setting content manager '
                 'to PloomberContentsManager')
    app.contents_manager_class = derive_class(app.contents_manager_class)

    try:
        # And re-run selected init steps from:
        # https://github.com/jupyter/notebook/blob/
        # 132f27306522b32fa667a6b208034cb7a04025c9/notebook/notebookapp.py#L1634-L1638
        app.contents_manager = app.contents_manager_class(parent=app,
                                                          log=app.log)
        app.session_manager.contents_manager = app.contents_manager
        app.web_app.settings["contents_manager"] = app.contents_manager
    except Exception:
        error = """[Ploomber] An error occured. Please
deactivate the server extension with "jupyter serverextension disable ploomber"
and configure the contents manager manually by adding
c.NotebookApp.contents_manager_class = "ploomber.jupyter.PloomberContentsManager"
to your .jupyter/jupyter_notebook_config.py file.
""" # noqa
        app.log.error(error)
        raise



from pathlib import Path


def _resolve(path):
    """
    Path.resolve() does not work on windows if the path doesn't exist
    this makes it work
    """
    path = Path(path)
    return path if path.is_absolute() else Path('.').resolve() / path



from ploomber.testing.sql.sqlparser import SQLParser
from ploomber.testing.sql.functions import (
    nulls_in_columns,
    distinct_values_in_column,
    range_in_column,
    exists_row_where,
)
from ploomber.testing.sql.duplicated import (
    duplicates_in_column,
    duplicates_stats,
    assert_no_duplicates_in_column,
)

__all__ = [
    'nulls_in_columns',
    'distinct_values_in_column',
    'range_in_column',
    'exists_row_where',
    'SQLParser',
    'duplicates_in_column',
    'duplicates_stats',
    'assert_no_duplicates_in_column',
]



"""
Functions to determine defaults

Layouts
-------
TODO: make sure soopervisor documentation also uses these two terms (simple and
packaged)
There are two layouts that can be used, simple and package.

Simple
------
A project composed of a pipeline.yaml and optionally extra pipeline.{name}.yaml
files (all of them must be in the same directory). the parent of pipeline.yaml
is considered the project root and relative sources in all pipeline yaml files
are so to the project root. Pipelines may be parametrized with env.{name}.yaml
files. {{here}} and {{root}} resolve to the same directory.

Packaged
--------
A packaged project has a setup.py file which is considered the project root. It
must contain a src/ directory with a single folder and pipelines are declared
in src/{package-name}/, pipeline.yaml is mandatory, with optional
pipeline.{name}.yaml. env.{name}.yaml can be siblings of pipeline.yaml but
env.yaml files in the current working working directory (which may not be
src/{package-name}) take precedence. {{here}} and {{root}} resolve to
different directories.

TODO: might be confusing for users that on this case, relative paths are so
to the project root (which is setup.py) instead of relative to pipeline.yaml,
which is under src/{package-name}/ how do we make this clear? same could be
say about errors when importing dotted paths (if the dotted path isn't in
full form and does not start with {package-name}). Maybe auto-detect if
we are inside a package and catch FileNotFoundError and ModuleNotFoundError
to provide guidance. e.g., check if the package is properly configured to
help with ModuleNotFoundError and when FileNotFoundError, modify the error
message to suggest using a {{here}} flag or a SourceLoader

TODO: make it clear that packaged projects should be installed with --editable
otherwise there isn't a notion of a setup.py in a parent directory.

TODO: auto-detect when the package is installed in site-packages but pipeline
is executed from local src/*, that should not be allows, recommend installing
in --editable mode

TODO: there is a unique where root_path is undefined: when a packge project
is called from site-packages (e.g ploomber build my_package::pipeline.yaml).
in such case the current working directory is set to be the root path

Default entry point
-------------------
To reduce typing, the command line interface and jupyter search for a
pipeline.yaml (not a pipeline.{name}.yaml) in standard locations. To override
these settings, users may set an environment variable to search for an
alternative file by setting the name (e.g., pipeline.serve.yaml) They have the
option to also use an environment variable to set the default env.{name}.yaml
to use. Note that only the basename is supplied not the path to the file,
since ploomber will look it up in standard locations. Although we could work
with the {name} portion alone (e.g., x in pipeline.x.yaml), we get the full
basename since it's more explicit.
"""
import warnings
import os
from glob import glob
from pathlib import Path
from os.path import relpath

from ploomber.exceptions import DAGSpecInvalidError
from ploomber.entrypoint import try_to_find_entry_point_type, EntryPoint


def _filesystem_root():
    return Path(os.path.abspath(os.sep))


def _package_location(root_path, name='pipeline.yaml'):
    """
    Look for a src/{package-name}/pipeline.yaml relative to root_path

    Parameters
    ----------
    root_path : str or pathlib.Path
        Looks for a package relative to this

    name : str, default='pipeline.yaml'
        YAML spec to search for

    Returns
    -------
    str
        Path to package. None if no package exists.
    """
    pattern = str(Path(root_path, 'src', '*', name))
    candidates = sorted([
        f for f in glob(pattern)
        if not str(Path(f).parent).endswith('.egg-info')
    ])

    if len(candidates) > 1:
        warnings.warn(f'Found more than one package location: {candidates}. '
                      f'Using the first one: {candidates[0]!r}')

    return candidates[0] if candidates else None


def entry_point_with_name(root_path=None, name=None):
    """Search for an entry point with a given name

    Parameters
    ----------
    name : str, default=None
        If None, searchs for a pipeline.yaml file otherwise for a
        file with such name
    """
    filename = name or 'pipeline.yaml'

    # first, find project root
    project_root = find_root_recursively(starting_dir=root_path,
                                         filename=filename)
    setup_py = Path(project_root, 'setup.py')
    setup_py_exists = setup_py.exists()

    # if ther is a setup.py file, look up a {project_root}/src/*/{name} file
    if setup_py_exists:
        entry_point = _package_location(root_path=project_root, name=filename)

        if entry_point is not None:
            return relpath(entry_point, Path().resolve())

    # otherwise use {project_root}/{file}. note that this file must
    # exust since find_root_recursively raises an error if it doesn't
    return relpath(Path(project_root, filename), Path().resolve())


# NOTE: this is described in doc/api/cli.rst. changes here must be also
# documented there
def entry_point(root_path=None):
    """
    Determines the default YAML specentry point. It first determines the
    project root. If the project isn't a package, it returns
    project_root/pipeline.yaml, otherwise src/*/pipeline.yaml. If the
    ENTRY_POINT environment variable is set, it looks for a file with
    such name (e.g., project_root/{ENTRY_POINT}).

    Parameters
    ----------
    root_path, optional
        Root path to look for the entry point. Defaults to the current working
        directory


    Notes
    -----
    Use cases for this function:
    * Called by the cli to locate the default entry point to use (nor args)
    * When deciding whether to add a new scaffold structure or parse the
        current one and add new files (catches DAGSpecInvalidError), no args

    Raises
    ------
    DAGSpecInvalidError
        If fails to determine project root or if no pipeline.yaml (or
        the content of the ENTRY_POINT environment variable, if any) exists in
        the expected location (once project root is determined).
    """
    # FIXME: rename env var used
    root_path = root_path or '.'
    env_var = os.environ.get('ENTRY_POINT')

    if env_var:
        if len(Path(env_var).parts) > 1:
            raise ValueError(f'ENTRY_POINT ({env_var!r}) '
                             'must be a filename and do not contain any '
                             'directory components (e.g., pipeline.yaml, '
                             'not path/to/pipeline.yaml).')

        filename = env_var
    else:
        filename = 'pipeline.yaml'

    return entry_point_with_name(root_path=root_path, name=filename)


def try_to_find_entry_point():
    """Try to find the default entry point. Returns None if it isn't possible
    """
    # check if it's a dotted path

    type_ = try_to_find_entry_point_type(os.environ.get('ENTRY_POINT'))

    if type_ == EntryPoint.DottedPath:
        return os.environ.get('ENTRY_POINT')

    # entry_point searches recursively for a YAML spec

    try:
        return entry_point(root_path=None)
    except Exception:
        # TODO: maybe display a warning with the error?
        pass


def entry_point_relative(name=None):
    """
    Returns a relative path to the entry point with the given name.

    Raises
    ------
    DAGSpecInvalidError
        If cannot locate the requested file

    Notes
    -----
    This is used by Soopervisor when loading dags. We must ensure it gets
    a relative path since such file is used for loading the spec in the client
    and in the hosted container (which has a different filesystem structure)
    """
    FILENAME = 'pipeline.yaml' if name is None else f'pipeline.{name}.yaml'

    location_pkg = _package_location(root_path='.', name=FILENAME)
    location = FILENAME if Path(FILENAME).exists() else None

    if location_pkg and location:
        raise DAGSpecInvalidError(
            f'Error loading {FILENAME}, both {location} '
            '(relative to the current working directory) '
            f'and {location_pkg} exist, but expected only one. '
            f'If your project is a package, keep {location_pkg}, '
            f'if it\'s not, keep {location}')

    if location_pkg is None and location is None:
        raise DAGSpecInvalidError(
            f'Could not find dag spec with name {FILENAME}, '
            'make sure the file is located relative to the working directory '
            f'or in src/pkg_name/{FILENAME} (where pkg_name is the name of '
            'your package if your project is one)')

    return location_pkg or location


def path_to_env_from_spec(path_to_spec):
    """

    It first looks up the PLOOMBER_ENV_FILENAME env var, it if exists, it uses
    the filename defined there. If not, it looks for an env.yaml file. Prefers
    a file in the working directory, otherwise, one relative to the spec's
    parent. If the appropriate env.yaml file does not exist, returns None.
    If path to spec has a pipeline.{name}.yaml format, it tries to look up an
    env.{name}.yaml first.

    It returns None if None of those files exist, except when
    PLOOMBER_ENV_FILENAME, in such case, it raises an error.

    Parameters
    ----------
    path_to_spec : str or pathlib.Path
        Path to YAML spec

    Raises
    ------
    FileNotFoundError
        If PLOOMBER_ENV_FILENAME is defined but doesn't exist
    ValueError
        If PLOOMBER_ENV_FILENAME is defined and contains a path with
        directory components.
        If path_to_spec does not have an extension or if it's a directory
    """
    # FIXME: delete this
    if path_to_spec is None:
        return None

    if Path(path_to_spec).is_dir():
        raise ValueError(
            f'Expected path to spec {str(path_to_spec)!r} to be a '
            'file but got a directory instead')

    if not Path(path_to_spec).suffix:
        raise ValueError('Expected path to spec to have a file extension '
                         f'but got: {str(path_to_spec)!r}')

    path_to_parent = Path(path_to_spec).parent
    environ = _get_env_filename_environment_variable(path_to_parent)
    if environ:
        filename = environ
    else:
        name = environ or extract_name(path_to_spec)
        filename = 'env.yaml' if name is None else f'env.{name}.yaml'

    return _search_for_env_with_name_and_parent(filename,
                                                path_to_parent,
                                                raise_=environ is not None)


def _search_for_env_with_name_and_parent(filename, path_to_parent, raise_):
    # pipeline.yaml....
    if filename is None:
        # look for env.yaml...
        return _path_to_filename_in_cwd_or_with_parent(
            filename='env.yaml', path_to_parent=path_to_parent, raise_=raise_)
    # pipeline.{name}.yaml
    else:
        # look for env.{name}.yaml
        path = _path_to_filename_in_cwd_or_with_parent(
            filename=filename, path_to_parent=path_to_parent, raise_=raise_)

        # not found, try with env.yaml...
        if path is None:
            return _path_to_filename_in_cwd_or_with_parent(
                filename='env.yaml',
                path_to_parent=path_to_parent,
                raise_=raise_)
        else:
            return path


def _path_to_filename_in_cwd_or_with_parent(filename, path_to_parent, raise_):
    """
    Looks for a file with filename in the current working directory, if it
    doesn't exist, it looks for it relative to path_to_parent.

    Parameters
    ----------
    filename : str
        Filename to search for

    path_to_parent : str or pathlib.Path
        If filename does not exist in the current working directory, look
        relative to this path

    raise_ : bool
        If Trye, raises an error if the file doesn't exist
    """
    local_env = Path('.', filename).resolve()

    if local_env.exists():
        return str(local_env)

    if path_to_parent:
        sibling_env = Path(path_to_parent, filename).resolve()

        if sibling_env.exists():
            return str(sibling_env)

    if raise_:
        raise FileNotFoundError('Failed to load env: PLOOMBER_ENV_FILENAME '
                                f'has value {filename!r} but '
                                'there isn\'t a file with such name. '
                                'Tried looking it up relative to the '
                                'current working directory '
                                f'({str(local_env)!r}) and relative '
                                f'to the YAML spec ({str(sibling_env)!r})')


def _get_env_filename_environment_variable(path_to_parent):
    env_var = os.environ.get('PLOOMBER_ENV_FILENAME')

    if env_var:
        if len(Path(env_var).parts) > 1:
            path_to_parent_abs = str(Path(path_to_parent).resolve())
            raise ValueError(f'PLOOMBER_ENV_FILENAME value ({env_var!r}) '
                             'must be a filename and do not contain any '
                             'directory components (e.g., env.yaml, '
                             'not path/to/env.yaml). Fix the value and place '
                             'the file in the same folder as the YAML '
                             f'spec ({path_to_parent_abs!r}) or next to the '
                             'current working directory')

    return env_var


def extract_name(path):
    """
    Extract name from a path whose filename is something.{name}.{extension}.
    Returns none if the file doesn't follow the naming convention
    """
    name = Path(path).name
    parts = name.split('.')

    if len(parts) < 3:
        return None
    else:
        return parts[1]


def find_file_recursively(name, max_levels_up=6, starting_dir=None):
    """
    Find a file by looking into the current folder and parent folders,
    returns None if no file was found otherwise pathlib.Path to the file

    Parameters
    ----------
    name : str
        Filename

    Returns
    -------
    path : str
        Absolute path to the file
    levels : int
        How many levels up the file is located
    """
    current_dir = starting_dir or os.getcwd()
    current_dir = Path(current_dir).resolve()
    path_to_file = None
    levels = None

    for levels in range(max_levels_up):
        current_path = Path(current_dir, name)

        if current_path.exists():
            path_to_file = current_path.resolve()
            break

        current_dir = current_dir.parent

    return path_to_file, levels


def find_parent_of_file_recursively(name, max_levels_up=6, starting_dir=None):
    path, levels = find_file_recursively(name,
                                         max_levels_up=6,
                                         starting_dir=starting_dir)

    if path:
        return path.parent, levels

    return None, None


def find_root_recursively(starting_dir=None,
                          filename=None,
                          check_parents=True):
    """
    Finds a project root by looking recursively for pipeline.yaml or a setup.py
    file. Ignores pipeline.yaml if located in src/*/pipeline.yaml.

    Parameters
    ---------
    starting_dir : str or pathlib.Path
        The directory to start the search


    Raises
    ------
    DAGSpecInvalidError
        If fails to determine a valid project root
    """
    filename = filename or 'pipeline.yaml'

    if len(Path(filename).parts) > 1:
        raise ValueError(f'{filename!r} should be a filename '
                         '(e.g., pipeline.yaml), not a path '
                         '(e.g., path/to/pipeline.yaml)')

    root_by_setup, setup_levels = find_parent_of_file_recursively(
        'setup.py', max_levels_up=6, starting_dir=starting_dir)

    root_by_pipeline, pipeline_levels = find_parent_of_file_recursively(
        filename,
        max_levels_up=6,
        starting_dir=starting_dir,
    )

    if root_by_pipeline == _filesystem_root():
        raise DAGSpecInvalidError(
            f'{filename} cannot be in the filesystem root. '
            'Please add it inside a directory like '
            f'project-name/{filename}. ')

    root_found = None

    # use found pipeline.yaml if not in a src/*/pipeline.yaml structure when
    # there is no setup.py
    # OR
    # if the pipeline.yaml if closer to the starting_dir than the setup.py.
    # e.g., project/some/pipeline.yaml vs project/setup.py
    if root_by_pipeline and not root_by_setup:
        if root_by_pipeline.parents[0].name == 'src':
            pkg_portion = str(Path(*root_by_pipeline.parts[-2:]))
            raise DAGSpecInvalidError(
                'Invalid project layout. Found project root at '
                f'{root_by_pipeline}  under a parent with name {pkg_portion}. '
                'This suggests a package '
                'structure but no setup.py exists. If your project is a '
                'package create a setup.py file, otherwise rename the '
                'src directory')

        root_found = root_by_pipeline
    elif root_by_pipeline and root_by_setup:
        if (pipeline_levels < setup_levels
                and root_by_pipeline.parents[0].name != 'src'):
            root_found = root_by_pipeline

    if root_by_setup and (not root_by_pipeline
                          or setup_levels <= pipeline_levels
                          or root_by_pipeline.parents[0].name == 'src'):
        pipeline_yaml = Path(root_by_setup, filename)

        # pkg_location checks if there is a src/{package-name}/{filename}
        # e.g., src/my_pkg/pipeline.yaml
        pkg_location = _package_location(root_path=root_by_setup,
                                         name=filename)

        if not pkg_location and not pipeline_yaml.exists():
            raise DAGSpecInvalidError(
                'Failed to determine project root. Found '
                'a setup.py file at '
                f'{str(root_by_setup)!r} and expected '
                f'to find a {filename} file at '
                f'src/*/{filename} (relative to '
                'setup.py parent) but no such file was '
                'found')

        if pkg_location and pipeline_yaml.exists():
            pkg = Path(*Path(pkg_location).parts[-3:-1])
            example = str(pkg / 'pipeline.another.yaml')
            raise DAGSpecInvalidError(
                'Failed to determine project root: found '
                f'two {filename} files: {pkg_location} '
                f'and {pipeline_yaml}. To fix it, move '
                'and rename the second file '
                f'under {str(pkg)} (e.g., {example}) or move {pkg_location} '
                'to your root directory')

        root_found = root_by_setup

    if root_found:
        if check_parents:
            try:
                another_found = find_root_recursively(
                    starting_dir=Path(root_found).parent,
                    filename=filename,
                    check_parents=False)
            except DAGSpecInvalidError:
                pass
            else:
                warnings.warn(
                    f'Found project root with filename {filename!r} '
                    f'at {str(root_found)!r}, but '
                    'found another one in a parent directory '
                    f'({str(another_found)!r}). The former will be used. '
                    'Nested YAML specs are not recommended, '
                    'consider moving them to the same folder and '
                    'rename them (e.g., project/pipeline.yaml '
                    'and project/pipeline.serve.yaml) or store them '
                    'in separate folders (e.g., '
                    'project1/pipeline.yaml and '
                    'project2/pipeline.yaml')

        return root_found
    else:
        raise DAGSpecInvalidError('Failed to determine project root. Looked '
                                  'recursively for a setup.py or '
                                  f'{filename} in parent folders but none of '
                                  'those files exist')


def try_to_find_root_recursively(starting_dir=None):
    # TODO: display warning
    try:
        return find_root_recursively(starting_dir=starting_dir)
    except Exception:
        pass


def find_package_name(starting_dir=None):
    """
    Find package name for this project. Raises an error if it cannot determine
    a valid root path
    """
    root = find_root_recursively(starting_dir=starting_dir)
    pkg = _package_location(root_path=root)

    if not pkg:
        raise ValueError(
            'Could not find a valid package. Make sure '
            'there is a src/package-name/pipeline.yaml file relative '
            f'to your project root ({root})')

    return Path(pkg).parent.name



import inspect
import types
from functools import wraps
from collections.abc import Mapping

from ploomber.env.env import Env
from ploomber.env.envdict import EnvDict


def _get_function_name_w_module(fn):
    mod_name = inspect.getmodule(fn).__name__
    return '.'.join((mod_name, fn.__name__))


def _validate_and_modify_signature(fn):
    sig = inspect.signature(fn)

    if not len(sig.parameters):
        raise RuntimeError('Function "{}" does not take arguments, '
                           '@with_env decorated functions should '
                           'have env as their first artgument'.format(
                               fn.__name__))

    if list(sig.parameters.keys())[0] != 'env':
        raise RuntimeError('Function "{}" does not "env" as its first '
                           'argument, which is required to use the '
                           '@with_env decorator'.format(fn.__name__))

    for arg in list(sig.parameters.keys())[1:]:
        if arg.lower().startswith('env'):
            raise RuntimeError('Function "{}" has arguments '
                               'starting with "env". Only the '
                               'first one should start with "env"'.format(
                                   fn.__name__))

    # https://www.python.org/dev/peps/pep-0362/#examples
    new_sig = sig.replace(parameters=tuple(sig.parameters.values())[1:])
    fn.__signature__ = new_sig

    # TODO: check no arg in the function starts with env (other than env)


def load_env(fn):
    """
    A function decorated with @load_env will be called with the current
    environment in an env keyword argument

    """
    _validate_and_modify_signature(fn)

    @wraps(fn)
    def wrapper(*args, **kwargs):
        return fn(Env.load(), *args, **kwargs)

    return wrapper


def with_env(source):
    """
    A function decorated with @with_env that starts and environment during
    the execution of a function.

    Notes
    -----
    The first argument of a function decorated with @with_env must be named
    "env", the env will be passed automatically when calling the function. The
    original function's signature is edited.

    You can replace values in the environment, e.g. if you want to replace
    env.key.another, you can call the decorated function with:
    my_fn(env__key__another='my_new_value')

    The environment is resolved at import time, changes to the working
    directory will not affect initialization.

    Examples
    --------
    .. literalinclude:: ../../examples/short/with_env.py

    """
    def decorator(fn):
        _validate_and_modify_signature(fn)

        try:
            # FIXME: we should deprecate initializing from a decorator
            # with a dictionary, it isn't useful. leaving it for now
            if isinstance(source, Mapping):
                env_dict = EnvDict(source)
            else:
                # when the decorator is called without args, look for
                # 'env.yaml'
                env_dict = EnvDict.find(source or 'env.yaml')
        except Exception as e:
            raise RuntimeError('Failed to resolve environment using '
                               '@with_env decorator in function "{}". '
                               'Tried to call Env with argument: {}'.format(
                                   _get_function_name_w_module(fn),
                                   source)) from e

        fn._env_dict = env_dict

        @wraps(fn)
        def wrapper(*args, **kwargs):
            to_replace = {
                k: v
                for k, v in kwargs.items() if k.startswith('env__')
            }

            for key in to_replace.keys():
                kwargs.pop(key)

            env_dict_new = env_dict._replace_flatten_keys(to_replace)

            try:
                Env._init_from_decorator(env_dict_new,
                                         _get_function_name_w_module(fn))
            except Exception as e:
                current = Env.load()
                raise RuntimeError('Failed to initialize environment using '
                                   '@with_env decorator in function "{}". '
                                   'Current environment: {}'.format(
                                       _get_function_name_w_module(fn),
                                       repr(current))) from e

            Env._ref = _get_function_name_w_module(fn)

            try:
                res = fn(Env.load(), *args, **kwargs)
            except Exception as e:
                Env.end()
                raise e

            Env.end()

            return res

        return wrapper

    if isinstance(source, types.FunctionType):
        fn = source
        source = None
        return decorator(fn)

    return decorator



"""
On languages and kernels
------------------------
NotebookSource represents source code in a Jupyter notebook format (language
agnostic). Apart from .ipynb, we also support any other extension supported
by jupytext.

Given a notebook, we have to know which language it is written in to extract
upstream/product variables (though this only happens when the option of
extracting dependencies automatically is on), we also have to determine the
Jupyter kernel to use (this is always needed).

The unequivocal place to store this information is in the notebook metadata
section, but given that we advocate for the use of scripts (converted to
notebooks via jupytext), they most likely won't contain metadata (metadata
saving is turned off by default in jupytext), so we have to infer this
ourselves.

To make things more complex, jupytext adds its own metadata section but we are
ignoring that for now.

Given that there are many places where this information might be stored, we
have a few rules to automatically determine language and kernel given a
script/notebook.
"""
from functools import wraps
import ast
from pathlib import Path
import warnings
from contextlib import redirect_stdout
from io import StringIO
from copy import deepcopy

# papermill is importing a deprecated module from pyarrow
with warnings.catch_warnings():
    warnings.simplefilter('ignore', FutureWarning)
    from papermill.parameterize import parameterize_notebook

import nbformat
import jupytext
from jupytext import cli as jupytext_cli
from jupytext.formats import long_form_one_format, short_form_one_format
from jupytext.config import JupytextConfiguration
import parso

from ploomber.exceptions import (SourceInitializationError,
                                 MissingParametersCellError)
from ploomber.placeholders.placeholder import Placeholder
from ploomber.util import requires
from ploomber.sources.abc import Source
from ploomber.sources.nb_utils import find_cell_with_tag, find_cell_with_tags
from ploomber.static_analysis.extractors import extractor_class_for_language
from ploomber.static_analysis.pyflakes import check_notebook
from ploomber.sources import docstring
from ploomber.io import pretty_print


def _jupytext_fmt(primitive, extension):
    """
    Determine the jupytext fmt string to use based on the content and extension
    """

    if extension != 'ipynb':
        fmt, _ = jupytext.guess_format(primitive, f'.{extension}')
        fmt_final = f'{extension}:{fmt}'
    else:
        fmt_final = '.ipynb'

    return fmt_final


# TODO: we should unit test that this function is called, as opposed to vanilla
# .read_text
def _read_primitive(path):
    """
    We read using the UTF-8 instead of the default encoding since notebooks are
    always stored in UTF-8.

    We can see this in nbformat, which always reads as UTF-8:
    https://github.com/jupyter/nbformat/blob/df63593b64a15ee1c37b522973c39e8674f93c5b/nbformat/__init__.py#L125

    Scripts are a different story since they may have other encodings, however,
    modern editors have UTF-8 as default (example: VSCode
    https://docs.microsoft.com/en-us/powershell/scripting/dev-cross-plat/vscode/understanding-file-encoding?view=powershell-7.2#configuring-vs-code)
    so it's safer to use UTF-8 than the default encoding.

    jupytext already does this:
    https://github.com/mwouts/jupytext/issues/896
    """
    return Path(path).read_text(encoding='utf-8')


def _get_last_cell(nb):
    """
    Get last cell, ignores cells with empty source (unless the notebook only
    has one cell and it's empty)
    """
    # iterate in reverse order
    for idx in range(-1, -len(nb.cells) - 1, -1):
        cell = nb.cells[idx]

        # only return it if it has some code
        if cell['source'].strip():
            return cell

    # otherwise return the first cell
    return nb.cells[0]


def _get_cell_suggestion(nb):
    format_name = nb.metadata.get('jupytext', {}).get('text_representation',
                                                      {}).get('format_name')

    preamble = 'Add a new cell with your code'

    if format_name == 'light':
        message = f'{preamble}:\n' + """
# + tags=["parameters"]
# your parameters here...
# -

# +
# your code here...
# -
"""

    elif format_name == 'percent':
        message = f'{preamble}:\n' + """
# %% tags=["parameters"]
# your parameters here...

# %%
# your code here...
"""
    else:
        message = preamble + '.'

    return message


def requires_path(func):
    """
    Checks if NotebookSource instance was initialized from a file, raises
    an error if not
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):

        if self._path is None:
            raise ValueError(f'Cannot use {func.__name__!r} if notebook was '
                             'not initialized from a file')

        return func(self, *args, **kwargs)

    return wrapper


class NotebookSource(Source):
    """
    A source object representing a jupyter notebook (or any format supported
    by jupytext)

    Parameters
    ----------
    hot_reload : bool, optional
        Makes the notebook always read the file before rendering

    kernelspec_name : str, optional
        Which kernel to use for executing the notebook, it overrides any
        existing kernelspec metadata in the notebook. If the notebook does
        not have kernelspec info, this parameter is required. Defaults to None.
        To see which kernelspecs are available run "jupyter kernelspec list"

    check_if_kernel_installed : bool, optional
        Check if the kernel is installed during initization

    Notes
    -----
    The render method prepares the notebook for execution: it adds the
    parameters and it makes sure kernelspec is defined
    """
    @requires([
        'parso', 'pyflakes', 'jupytext', 'nbformat', 'papermill',
        'jupyter_client'
    ])
    def __init__(self,
                 primitive,
                 hot_reload=False,
                 ext_in=None,
                 kernelspec_name=None,
                 static_analysis='regular',
                 check_if_kernel_installed=True):
        # any non-py file must first be converted using jupytext, we need
        # that representation for validation, if input is already a .py file
        # do not convert. If passed a string, try to guess format using
        # jupytext. We also need ipynb representation for .develop(),
        # but do lazy loading in case we don't need both
        self._primitive = primitive
        self._check_if_kernel_installed = check_if_kernel_installed

        # this happens if using SourceLoader
        if isinstance(primitive, Placeholder):
            self._path = primitive.path
            self._primitive = str(primitive)
        elif isinstance(primitive, str):
            self._path = None
            self._primitive = primitive
        elif isinstance(primitive, Path):
            self._path = primitive

            if primitive.is_dir():
                raise SourceInitializationError(
                    f'Failed to initialize {str(primitive)!r}. '
                    'Expected a file, got a directory.' +
                    _suggest_ploomber_scaffold_is_dir())

            if not primitive.exists():
                raise SourceInitializationError(
                    f'Failed to initialize {str(primitive)!r}. '
                    'File does not exist.' +
                    _suggest_ploomber_scaffold_missing_file())

            self._primitive = _read_primitive(primitive)
        else:
            raise TypeError('Notebooks must be initialized from strings, '
                            'Placeholder or pathlib.Path, got {}'.format(
                                type(primitive)))

        static_analysis_vals = {'disable', 'regular', 'strict'}

        if static_analysis not in static_analysis_vals:
            raise ValueError(f'{static_analysis!r} is not a '
                             "valid 'static_analysis' value, choose one from: "
                             f'{pretty_print.iterable(static_analysis_vals)}')

        self.static_analysis = static_analysis
        self._kernelspec_name = kernelspec_name
        self._hot_reload = hot_reload

        # TODO: validate ext_in values and extensions

        if self._path is None and hot_reload:
            raise ValueError('hot_reload only works in the notebook was '
                             'loaded from a file')

        if self._path is not None and ext_in is None:
            self._ext_in = self._path.suffix[1:]
        elif self._path is None and ext_in is None:

            if Path(self._primitive).exists():
                path = str(self._primitive)
                raise ValueError(
                    f'The file {path!r} you passed looks like '
                    'a path to a file. Perhaps you meant passing a '
                    'pathlib.Path object? Example:\n\n'
                    'from pathlib import Path\n'
                    f'NotebookRunner(Path({path!r}))')

            else:
                raise ValueError(
                    '"ext_in" cannot be None if the notebook is '
                    'initialized from a string. Either pass '
                    'a pathlib.Path object with the notebook file '
                    'location or pass the source code as string '
                    'and include the "ext_in" parameter')

        elif self._path is not None and ext_in is not None:
            raise ValueError('"ext_in" must be None if notebook is '
                             'initialized from a pathlib.Path object')
        elif self._path is None and ext_in is not None:
            self._ext_in = ext_in

        # try to determine language based on extension, though this test
        # might be inconclusive if dealing with a ipynb file, though we only
        # use this to determine the appropriate jupyter kernel when
        # initializing from a string, when initializing from files, the
        # extension is used to determine the kernel
        self._language = determine_language(self._ext_in)

        self._loc = None
        self._params = None

        self._nb_str_unrendered = None
        self._nb_obj_unrendered = None
        self._nb_str_rendered = None
        self._nb_obj_rendered = None

        # this will raise an error if kernelspec_name is invalid
        self._read_nb_str_unrendered()

        self._post_init_validation(str(self._primitive))

    @property
    def primitive(self):
        if self._hot_reload:
            self._primitive = _read_primitive(self._path)

        return self._primitive

    def render(self, params):
        """Render notebook (fill parameters using papermill)
        """
        self._params = json_serializable_params(params)
        self._render()

    def _render(self):
        # _read_nb_str_unrendered uses hot_reload, this ensures we always get
        # the latest version
        _, nb = self._read_nb_str_unrendered()

        if 'parameters' in _get_last_cell(nb).metadata.get('tags', []):
            cell_suggestion = _get_cell_suggestion(nb)
            kind = 'notebook' if self._ext_in == 'ipynb' else 'script'
            raise SourceInitializationError(
                f'Error processing {str(self._path)!r}: the last cell '
                f'in the {kind} is the parameters cell. {cell_suggestion}')

        # this is needed for parameterize_notebook to work
        for cell in nb.cells:
            if not hasattr(cell.metadata, 'tags'):
                cell.metadata['tags'] = []
        nb.metadata['papermill'] = dict()

        # NOTE: we use parameterize_notebook instead of execute_notebook
        # with the prepare_only option because the latter adds a "papermill"
        # section on each cell's metadata, which makes it too verbose when
        # using NotebookRunner.develop() when the source is script (each cell
        # will have an empty "papermill" metadata dictionary)
        nb = parameterize_notebook(nb, self._params)

        # delete empty tags to prevent cluttering the notebooks
        for cell in nb.cells:
            if not len(cell.metadata['tags']):
                cell.metadata.pop('tags')

        self._nb_str_rendered = nbformat.writes(nb)
        self._post_render_validation()

    def _read_nb_str_unrendered(self):
        """
        Returns the notebook representation (JSON string), this is the raw
        source code passed, does not contain injected parameters.

        Adds kernelspec info if not present based on the kernelspec_name,
        this metadata is required for papermill to know which kernel to use.

        An exception is raised if we cannot determine kernel information.
        """
        # hot_reload causes to always re-evalaute the notebook representation
        if self._nb_str_unrendered is None or self._hot_reload:
            # this is the notebook node representation
            nb = _to_nb_obj(
                self.primitive,
                ext=self._ext_in,
                # passing the underscored version
                # because that's the only one available
                # when this is initialized
                language=self._language,
                kernelspec_name=self._kernelspec_name,
                check_if_kernel_installed=self._check_if_kernel_installed,
                path=self._path)

            # if the user injected cells manually (with ploomber nb --inject)
            # the source will contain the injected cell, remove it because
            # it should not be considered part of the source code
            self._nb_obj_unrendered = _cleanup_rendered_nb(nb, print_=False)

            # get the str representation. always write from nb_obj, even if
            # this was initialized with a ipynb file, nb_obj contains
            # kernelspec info
            self._nb_str_unrendered = nbformat.writes(
                self._nb_obj_unrendered, version=nbformat.NO_CONVERT)

        return self._nb_str_unrendered, self._nb_obj_unrendered

    def _post_init_validation(self, value):
        """
        Validate notebook after initialization (run pyflakes to detect
        syntax errors)
        """
        # NOTE: what happens if I pass source code with errors to parso?
        # maybe we don't need to use pyflakes after all
        # we can also use compile. can pyflakes detect things that
        # compile cannot?
        params_cell, _ = find_cell_with_tag(self._nb_obj_unrendered,
                                            'parameters')

        if params_cell is None:
            loc = ' "{}"'.format(self.loc) if self.loc else ''
            msg = ('Notebook{} does not have a cell tagged '
                   '"parameters"'.format(loc))

            if self.loc and Path(self.loc).suffix == '.py':
                msg += """.
Add a cell at the top like this:

# %% tags=["parameters"]
upstream = None
product = None

Go to: https://ploomber.io/s/params for more information
"""
            if self.loc and Path(self.loc).suffix == '.ipynb':
                msg += ('. Add a cell at the top and tag it as "parameters". '
                        'Go to the next URL for '
                        'details: https://ploomber.io/s/params')

            raise MissingParametersCellError(msg)

    def _post_render_validation(self):
        """
        Validate params passed against parameters in the notebook
        """
        # NOTE: maybe static_analysis = off should not turn off everything
        # but only warn

        # strict mode: raise and check signature
        # regular mode: _check_notebook called in NotebookRunner.run
        if self.static_analysis == 'strict':
            self._check_notebook(raise_=True, check_signature=True)
        else:
            # otherwise, only warn on unused parameters
            _warn_on_unused_params(self._nb_obj_unrendered, self._params)

    def _check_notebook(self, raise_, check_signature):
        if self.static_analysis and self.language == 'python':
            # warn if errors (e.g., undeclared variables, syntax errors)
            check_notebook(self._nb_str_to_obj(self._nb_str_rendered),
                           self._params,
                           filename=self._path or 'notebook',
                           raise_=raise_,
                           check_signature=check_signature)

    @property
    def doc(self):
        """
        Returns notebook docstring parsed either from a triple quoted string
        in the top cell or a top markdown markdown cell
        """
        return docstring.extract_from_nb(self._nb_obj_unrendered)

    @property
    def loc(self):
        return self._path

    @property
    def name(self):
        # filename without extension(e.g., plot.py -> plot)
        if self._path:
            return self._path.stem

    @property
    def nb_str_rendered(self):
        """
        Returns the notebook (as a string) with parameters injected, hot
        reloadig if necessary
        """
        if self._nb_str_rendered is None:
            raise RuntimeError('Attempted to get location for an unrendered '
                               'notebook, render it first')

        if self._hot_reload:
            self._render()

        return self._nb_str_rendered

    @property
    def nb_obj_rendered(self):
        """
        Returns the notebook (as an objet) with parameters injected, hot
        reloadig if necessary
        """
        if self._nb_obj_rendered is None:
            # using self.nb_str_rendered triggers hot reload if needed
            self._nb_obj_rendered = self._nb_str_to_obj(self.nb_str_rendered)

        return self._nb_obj_rendered

    def __str__(self):
        # reload if empty or hot_reload=True
        self._read_nb_str_unrendered()

        # FIXME: this should ignore changes to the markdown cells
        return '\n'.join([c.source for c in self._nb_obj_unrendered.cells])

    def __repr__(self):
        if self.loc is not None:
            return "{}('{}')".format(type(self).__name__, self.loc)
        else:
            return "{}(loaded from string)".format(type(self).__name__)

    @property
    def variables(self):
        raise NotImplementedError

    @property
    def extension(self):
        # this can be Python, R, Julia, etc. We are handling them the same,
        # for now, no normalization can be done.
        # One approach is to use the ext if loaded from file, otherwise None
        return None

    # FIXME: add this to the abstract class, probably get rid of "extension"
    # since it's not informative (ipynb files can be Python, R, etc)
    @property
    def language(self):
        """
        Notebook Language (Python, R, etc), this is a best-effort property,
        can be None if we could not determine the language
        """
        if self._language is None:
            self._read_nb_str_unrendered()

            try:
                # make sure you return "r" instead of "R"
                return (self._nb_obj_unrendered.metadata.kernelspec.language.
                        lower())
            except AttributeError:
                return None

        else:
            return self._language

    def _nb_str_to_obj(self, nb_str):
        return nbformat.reads(nb_str, as_version=nbformat.NO_CONVERT)

    def _get_parameters_cell(self):
        self._read_nb_str_unrendered()
        cell, _ = find_cell_with_tag(self._nb_obj_unrendered, tag='parameters')
        return cell.source

    def extract_upstream(self):
        extractor_class = extractor_class_for_language(self.language)
        return extractor_class(self._get_parameters_cell()).extract_upstream()

    def extract_product(self):
        extractor_class = extractor_class_for_language(self.language)
        return extractor_class(self._get_parameters_cell()).extract_product()

    @requires_path
    def save_injected_cell(self):
        """
        Inject cell, overwrite the source file (and any paired files)
        """
        fmt_ = _jupytext_fmt(self._primitive, self._ext_in)

        # add metadata to flag that the cell was injected manually
        recursive_update(
            self.nb_obj_rendered,
            dict(metadata=dict(ploomber=dict(injected_manually=True))))

        # Are we updating a text file that has a metadata filter? If so,
        # add ploomber as a section that must be stored
        if (self.nb_obj_rendered.metadata.get(
                'jupytext', {}).get('notebook_metadata_filter') == '-all'):
            recursive_update(
                self.nb_obj_rendered,
                dict(metadata=dict(jupytext=dict(
                    notebook_metadata_filter='ploomber,-all'))))

        # overwrite
        jupytext.write(self.nb_obj_rendered, self._path, fmt=fmt_)

        # overwrite all paired files
        for path, fmt_ in iter_paired_notebooks(self.nb_obj_rendered, fmt_,
                                                self._path.stem):
            jupytext.write(self.nb_obj_rendered, fp=path, fmt=fmt_)

    @requires_path
    def remove_injected_cell(self):
        """
        Delete injected cell, overwrite the source file (and any paired files)
        """
        nb_clean = _cleanup_rendered_nb(self._nb_obj_unrendered)

        # remove metadata
        recursive_update(
            nb_clean,
            dict(metadata=dict(ploomber=dict(injected_manually=None))))

        fmt_ = _jupytext_fmt(self._primitive, self._ext_in)

        # overwrite
        jupytext.write(nb_clean, self._path, fmt=fmt_)

        # overwrite all paired files
        for path, fmt_ in iter_paired_notebooks(self._nb_obj_unrendered, fmt_,
                                                self._path.stem):
            jupytext.write(nb_clean, fp=path, fmt=fmt_)

    @requires_path
    def format(self, fmt):
        """Change source format

        Returns
        -------
        str
            The path if the extension changed, None otherwise
        """
        nb_clean = _cleanup_rendered_nb(self._nb_obj_unrendered)

        ext_file = self._path.suffix
        ext_format = long_form_one_format(fmt)['extension']
        extension_changed = ext_file != ext_format

        if extension_changed:
            path = self._path.with_suffix(ext_format)
            Path(self._path).unlink()
        else:
            path = self._path

        jupytext.write(nb_clean, path, fmt=fmt)

        return path if extension_changed else None

    @requires_path
    def pair(self, base_path):
        """Pairs with an ipynb file
        """
        # TODO: add unit test
        if self._ext_in == 'ipynb':
            raise ValueError(
                'pairing only works with .py files, got .ipynb. '
                'Yoy may convert the .ipynb to .py and try again.')

        fmt, _ = jupytext.guess_format(self._primitive, f'.{self._ext_in}')
        fmt_ = f'{self._ext_in}:{fmt}'

        # mute jupytext's output
        with redirect_stdout(StringIO()):
            jupytext_cli.jupytext(args=[
                '--set-formats', f'{base_path}//ipynb,{fmt_}',
                str(self._path)
            ])

    @requires_path
    def sync(self):
        """Pairs with and ipynb file
        """
        # mute jupytext's output
        with redirect_stdout(StringIO()):
            jupytext_cli.jupytext(args=['--sync', str(self._path)])


def json_serializable_params(params):
    # papermill only allows JSON serializable parameters
    # convert Params object to dict
    params = params.to_dict()
    params['product'] = params['product'].to_json_serializable()

    if params.get('upstream'):
        params['upstream'] = params['upstream'].to_json_serializable()

    return params


def _to_nb_obj(source,
               language,
               ext=None,
               kernelspec_name=None,
               check_if_kernel_installed=True,
               path=None):
    """
    Convert to jupyter notebook via jupytext, if the notebook does not contain
    kernel information and the user did not pass a kernelspec_name explicitly,
    we will try to infer the language and select a kernel appropriately.

    If a valid kernel is found, it is added to the notebook. If none of this
    works, an exception is raised.

    If also converts the code string to its notebook node representation,
    adding kernel data accordingly.

    Parameters
    ----------
    source : str
        Jupyter notebook (or jupytext compatible formatted) document

    language : str
        Programming language

    path : str, default=None
        Script/notebook path. If not None, it's used to throw an informative
        error if the notebook fails to load

    Returns
    -------
    nb
        Notebook object


    Raises
    ------
    RenderError
        If the notebook has no kernelspec metadata and kernelspec_name is
        None. A notebook without kernelspec metadata will not display in
        jupyter notebook correctly. We have to make sure all notebooks
        have this.
    """
    import jupytext

    # let jupytext figure out the format
    try:
        nb = jupytext.reads(source, fmt=ext)
    except Exception as e:
        what = 'notebook' if ext == 'ipynb' else 'script'
        err = f'Failed to read {what}'

        if path is not None:
            err += f' from {str(path)!r}'

        raise SourceInitializationError(err) from e

    # NOTE: I can add the cell with parameters here, but what happens if
    # extract_upstream is false? would that be a problem?

    check_nb_kernelspec_info(nb,
                             kernelspec_name,
                             ext,
                             language,
                             check_if_installed=check_if_kernel_installed)

    return nb


def check_nb_kernelspec_info(nb,
                             kernelspec_name,
                             ext,
                             language,
                             check_if_installed=True):
    """Make sure the passed notebook has kernel info

    Parameters
    ----------
    check_if_installed : bool
        Also check if the kernelspec is installed, nb.metadata.kernelspec
        to be replaced by whatever information jupyter returns when requesting
        the kernelspec
    """
    import jupyter_client

    kernel_name = determine_kernel_name(nb, kernelspec_name, ext, language)

    # cannot keep going if we don't have the kernel name
    if kernel_name is None:
        raise SourceInitializationError(
            'Notebook does not contain kernelspec metadata and '
            'kernelspec_name was not specified, either add '
            'kernelspec info to your source file or specify '
            'a kernelspec by name. To see list of installed kernels run '
            '"jupyter kernelspec list" in the terminal (first column '
            'indicates the name). Python is usually named "python3", '
            'R usually "ir"')

    if check_if_installed:
        kernelspec = jupyter_client.kernelspec.get_kernel_spec(kernel_name)

        nb.metadata.kernelspec = {
            "display_name": kernelspec.display_name,
            "language": kernelspec.language,
            "name": kernel_name
        }
    else:
        if 'metadata' not in nb:
            nb['metadata'] = dict()

        if 'kernelspec' not in nb['metadata']:
            nb['metadata']['kernelspec'] = dict()

        # we cannot ask jupyter, so we fill this in ourselves
        nb.metadata.kernelspec = {
            "display_name": 'R' if kernel_name == 'ir' else 'Python 3',
            "language": 'R' if kernel_name == 'ir' else 'python',
            "name": kernel_name
        }


def determine_kernel_name(nb, kernelspec_name, ext, language):
    """
    Determines the kernel name by using the following data (returns whatever
    gives kernel info first): 1) explicit kernel from the user 2) notebook's
    metadata 3) file extension 4) language 5) best guess
    """
    # explicit kernelspec name
    if kernelspec_name is not None:
        return kernelspec_name

    # use metadata info
    try:
        return nb.metadata.kernelspec.name
    except AttributeError:
        pass

    # use language from extension if passed, otherwise use language variable
    if ext:
        language = determine_language(ext)

    lang2kernel = {'python': 'python3', 'r': 'ir'}

    if language in lang2kernel:
        return lang2kernel[language]

    # nothing worked, try to guess if it's python...
    is_python_ = is_python(nb)

    if is_python_:
        return 'python3'
    else:
        return None


def inject_cell(model, params):
    """Inject params (by adding a new cell) to a model

    Notes
    -----
    A model is different than a notebook:
    https://jupyter-notebook.readthedocs.io/en/stable/extending/contents.html
    """
    nb = nbformat.from_dict(model['content'])

    # we must ensure nb has kernelspec info, otherwise papermill will fail to
    # parametrize
    ext = model['name'].split('.')[-1]
    check_nb_kernelspec_info(nb, kernelspec_name=None, ext=ext, language=None)

    # papermill adds a bunch of things before calling parameterize_notebook
    # if we don't add those things, parameterize_notebook breaks
    # https://github.com/nteract/papermill/blob/0532d499e13e93d8990211be33e9593f1bffbe6c/papermill/iorw.py#L400
    if not hasattr(nb.metadata, 'papermill'):
        nb.metadata['papermill'] = {
            'parameters': dict(),
            'environment_variables': dict(),
            'version': None,
        }

    for cell in nb.cells:
        if not hasattr(cell.metadata, 'tags'):
            cell.metadata['tags'] = []

    params = json_serializable_params(params)

    comment = ('This cell was injected automatically based on your stated '
               'upstream dependencies (cell above) and pipeline.yaml '
               'preferences. It is temporary and will be removed when you '
               'save this notebook')

    model['content'] = parameterize_notebook(nb,
                                             params,
                                             report_mode=False,
                                             comment=comment)


def _cleanup_rendered_nb(nb, print_=True):
    """
    Cleans up a rendered notebook object. Removes cells with tags:
    injected-parameters, debugging-settings, and metadata injected by
    papermill
    """
    out = find_cell_with_tags(nb,
                              ['injected-parameters', 'debugging-settings'])

    if print_:
        for key in out.keys():
            print(f'Removing {key} cell...')

    idxs = set(cell['index'] for cell in out.values())

    nb['cells'] = [
        cell for idx, cell in enumerate(nb['cells']) if idx not in idxs
    ]

    # papermill adds "tags" to all cells that don't have them, remove them
    # if they are empty to avoid cluttering the script
    for cell in nb['cells']:
        if 'tags' in cell.get('metadata', {}):
            if not len(cell['metadata']['tags']):
                del cell['metadata']['tags']

    return nb


def is_python(nb):
    """
    Determine if the notebook is Python code for a given notebook object, look
    for metadata.kernelspec.language first, if not defined, try to guess if
    it's Python, it's conservative and it returns False if the code is valid
    Python but contains (<-), in which case it's much more likely to be R
    """
    is_python_ = None

    # check metadata first
    try:
        language = nb.metadata.kernelspec.language
    except AttributeError:
        pass
    else:
        is_python_ = language == 'python'

    # no language defined in metadata, check if it's valid python
    if is_python_ is None:
        code_str = '\n'.join([c.source for c in nb.cells])

        try:
            ast.parse(code_str)
        except SyntaxError:
            is_python_ = False
        else:
            # there is a lot of R code which is also valid Python code! So
            # let's
            # run a quick test. It is very unlikely to have "<-" in Python (
            # {less than} {negative} but extremely common {assignment}
            if '<-' not in code_str:
                is_python_ = True

    # inconclusive test...
    if is_python_ is None:
        is_python_ = False

    return is_python_


def determine_language(extension):
    """
    A function to determine programming language given file extension,
    returns programming language name (all lowercase) if could be determined,
    None if the test is inconclusive
    """
    if extension.startswith('.'):
        extension = extension[1:]

    mapping = {'py': 'python', 'r': 'r', 'R': 'r', 'Rmd': 'r', 'rmd': 'r'}

    # ipynb can be many languages, it must return None
    return mapping.get(extension)


def recursive_update(target, update):
    """Recursively update a dictionary. Taken from jupytext.header
    """
    for key in update:
        value = update[key]
        if value is None:
            # remove if it exists
            target.pop(key, None)
        elif isinstance(value, dict):
            target[key] = recursive_update(target.get(key, {}), value)
        else:
            target[key] = value
    return target


def parse_jupytext_format(fmt, name):
    """
    Parse a jupytext format string (such as notebooks//ipynb) and return the
    path to the file and the extension
    """
    fmt_parsed = long_form_one_format(fmt)
    path = Path(fmt_parsed['prefix'], f'{name}{fmt_parsed["extension"]}')
    del fmt_parsed['prefix']
    return path, short_form_one_format(fmt_parsed)


def iter_paired_notebooks(nb, fmt_, name):
    formats = nb.metadata.get('jupytext', {}).get('formats', '')

    if not formats:
        return

    formats = formats.split(',')

    formats.remove(fmt_)

    # overwrite all paired files
    for path, fmt_current in (parse_jupytext_format(fmt, name)
                              for fmt in formats):
        yield path, fmt_current


def _nb2codestr(nb):
    return '\n'.join([c.source for c in nb.cells if c.cell_type == 'code'])


def _warn_on_unused_params(nb, params):
    nb = deepcopy(nb)
    _, idx = find_cell_with_tag(nb, 'parameters')
    del nb.cells[idx]

    code = _nb2codestr(nb)

    # NOTE: if there a syntax error we cannot accurately check this
    m = parso.parse(code)
    names = set(m.get_used_names())

    # remove product since it may not be required
    # FIXME: maybe only remove it if it's a dictionary with >2 keys
    unused = set(params) - names - {'product'}

    if unused:
        warnings.warn('These parameters are not used in the '
                      f'task\'s source code: {pretty_print.iterable(unused)}')


def add_parameters_cell(path, extract_upstream, extract_product):
    """
    Add parameters cell to a script/notebook in the given path, overwrites the
    original file
    """
    source = ''

    if extract_upstream:
        source += """\
# declare a list tasks whose products you want to use as inputs
upstream = None
"""
    if extract_product:
        source += """\
# declare a dictionary with the outputs of this task
product = None
"""

    c = JupytextConfiguration()
    c.notebook_metadata_filter
    c.cell_metadata_filter = 'all'

    nb = jupytext.read(path)
    new_cell = nbformat.v4.new_code_cell(source,
                                         metadata={'tags': ['parameters']})
    nb.cells.insert(0, new_cell)
    jupytext.write(nb, path, config=c)


def _suggest_ploomber_scaffold_missing_file():
    if Path('pipeline.yaml').is_file():
        return '\nTo create it, run: ploomber scaffold'
    else:
        return ''


def _suggest_ploomber_scaffold_is_dir():
    if Path('pipeline.yaml').is_file():
        return ('\nTo create it, delete the directory, '
                'then run: ploomber scaffold')
    else:
        return ''
